"""
ECCO v4 Python: read_bin_llc

This module includes utility routines for loading binary files in the
llc 13-tile native flat binary layout.  This layout is the default for
MITgcm input and output for global setups using lat-lon-cap (llc) layout.
The llc layout is used for ECCO v4.

.. _ecco_v4_py Documentation :
   https://github.com/ECCO-GROUP/ECCOv4-py

"""

from __future__ import division,print_function
from xmitgcm import open_mdsdataset
from xmitgcm.file_utils import clear_cache
import xmitgcm
import numpy as np
import xarray as xr
import time
import sys
import datetime

from .llc_array_conversion  import llc_compact_to_tiles, \
    llc_compact_to_faces, llc_faces_to_tiles, llc_faces_to_compact, \
    llc_tiles_to_faces, llc_tiles_to_compact

from .read_bin_gen import load_binary_array

from .ecco_utils import make_time_bounds_and_center_times_from_ecco_dataset
from .ecco_utils import add_global_metadata
from .ecco_utils import add_variable_metadata
from .ecco_utils import add_coordinate_metadata
from .ecco_utils import sort_all_attrs, sort_attrs
from .ecco_utils import extract_yyyy_mm_dd_hh_mm_ss_from_datetime64 

def load_ecco_vars_from_mds(mds_var_dir,
                            mds_grid_dir=None,
                            mds_files=None,
                            vars_to_load = 'all',
                            tiles_to_load = [0,1,2,3,4,5,6,7,8,9,10,11,12],
                            model_time_steps_to_load = 'all',
                            model_start_datetime = np.datetime64('1992-01-01T12:00:00'),
                            delta_t = 3600,
                            output_freq_code = '',
                            drop_unused_coords = False,
                            grid_vars_to_coords = True,
                            coordinate_metadata = [],
                            variable_metadata = [],
                            global_metadata = [],
                            cell_bounds = None,
                            mds_datatype = '>f4',
                            llc_method = 'bigchunks',
                            less_output=True,
                            read_grid=True,
                            **kwargs):

    """

    Uses xmitgcm's *open_mdsdataset* routine to load ecco variable(s) from
    MITgcm's MDS binary output into xarray Dataset/DataArray objects.

    The main benefit of using this routine over open_mdsdataset is that this
    routine allows for

        - proper centering of the *time* variable for time-averaged fields

        - creation of the *time-bnds* fields in time-averaged fields

        - specification of extra variable-specific and globl metadata


    xmitgcm.open_mdsdataset uses the model step number from the file name
    (e.g., 732 from the file VAR.000000000732.data) to construct the
    'time' field.  For time-averaged fields, this model step
    corresponds to END of the averaging period, not the time averaging mid
    point. This routine fixes the 'time' field of time-averaged fields
    to be the mid point of the time averaging period when the appropriate
    *output_freq_code* is passed.

    Parameters
    ----------
    mds_var_dir : str
        directory where the .data/.meta files are stored

    mds_grid_dir : str, optional, default 'None'
        the directory where the model binary (.data) grid fields
        are stored, default is same directory as mds_var_dir

    mds_files   : str or list or None, optional
        either: a string or list of file names to load,
        or None to load all files
        Note :  the name is everything BEFORE the time step
        the mds_file name for 'var.000000000732.data' is 'var'

    vars_to_load : str or list, optional, default 'all'
        a string or list of the variable names to read from the mds_files

        - if 'all' then all variables in the files are loaded

    tiles_to_load : int or list of ints, optional, default range(13)
        an int or list of ints indicating which tiles to load

    model_time_steps_to_load : int or list of ints, optional, default 'all'
        an int or list of ints indicating which model time steps to load

        Note : the model time step indicates the time step when the the file was written.
        when the field is a time average, this time step shows the END of the averaging period.


    model_start_datetime : numpy datetime64 object, optional, default: np.datetime64('1992-01-01T12:00:00')
        a numpy datetime64 object with the time set to the beginning of the simulation. 
        useful to change if loading an extension to the main solution where the 
        output field iteration numbers have reset to zero.

        default is 1992-01-01T12:00:00, the value used in ECCO V4. 

    delta_t : int, optional, default 3600 (seconds) 
        the length of the model time step, in seconds. V4 uses 1 hour (3600s) time steps 
        useful to change if re-running the model with a different time step
        Note: a different time step will change the solution 

    output_freq_code : str, optional, default empty string
        a code used to create the proper time indices on the fields after loading
        ('AVG' or 'SNAPSHOT') + '_' + ('DAY','WEEK','MON', or 'YEAR')

        valid options :
        - AVG_DAY, AVG_WEEK, AVG_MON, AVG_YEAR
        - SNAPSHOT_DAY, SNAPSHOT_WEEK, SNAPSHOT_MON, SNAPSHOT_YEAR

    drop_unused_coords : boolean, optional, default False
        drop coordinates that have dimensions that do not appear in any
        data variables

    grid_vars_to_coords : boolean, optional, default True
        promote grid variables to coordinates. grid variables are coordinates
        according to xarray's definition but not CF conventions

    coordinate_metadata : list, option, default empty list
        tuples with information that is specific to coordinate fields

    variable_metadata : list, option, default empty list
        tuples with information specific to variable fields

    global_metadata : list, option, default empty list
        tuples with global metadata information

    cell_bounds: a DataSet object, optional, default None
        DataSet object with two DataArrays, XC_bnds, YC_bnds
        If passed, the auxillary XC_bnds and YC_bnds coordinates are
        added

    mds_datatype : string, optional, default '>f4'
        code indicating what type of field to load if the xmitgcm cannot
        determine the format from the .meta file.  '>f4' means big endian
        32 bit float.

    llc_method : string, optional, default 'big_chunks'
        refer to the xmitgcm documentation.

    less_output : logical, optional
        if True (default), omit additional print statements

    **kwargs: optional
        extra inputs passed to xmitgcm.open_mdsdataset

    Returns
    =======

    ecco_dataset : xarray Dataset


    """

    # range object is different between python 2 and 3
    if sys.version_info[0] >= 3 and isinstance(tiles_to_load, range):
        tiles_to_load = list(tiles_to_load)

    #ECCO v4 r3 starts 1992/1/1 12:00:00
    #start_year = 1992
    #start_mon  = 1
    #start_day  = 1
    #start_hour = 12
    #start_min  = 0
    #start_sec  = 0

    start_year, start_mon, start_day, start_hour, start_min, start_sec = \
      extract_yyyy_mm_dd_hh_mm_ss_from_datetime64(model_start_datetime)
 
    # define reference date for xmitgcm
    ref_date = str(start_year) + '-' + str(start_mon)  + '-' + \
        str(start_day)  + ' ' +  str(start_hour) + ':' +  \
        str(start_min)  + ':' + str(start_sec)

    # clear cache so listdir() will find the newly downloaded mds files for the current timestep
    clear_cache()

    if model_time_steps_to_load == 'all':
        if not less_output:
            print ('loading all model time steps')
            print('read bin_llc:')
            print(mds_var_dir)
            print(mds_grid_dir)

        ecco_dataset = open_mdsdataset(data_dir = mds_var_dir,
                                       grid_dir = mds_grid_dir,
                                       read_grid = read_grid,
                                       prefix = mds_files,
                                       geometry = 'llc',
                                       iters = 'all',
                                       ref_date = ref_date,
                                       delta_t  = delta_t,
                                       default_dtype = np.dtype(mds_datatype),
                                       grid_vars_to_coords=grid_vars_to_coords,
                                       llc_method = llc_method,
                                       ignore_unknown_vars=False,
                                       **kwargs)

    else:
        if not less_output:
            print ('loading subset of  model time steps')

        if isinstance(model_time_steps_to_load, int):
            model_time_steps_to_load = [model_time_steps_to_load]

        if isinstance(model_time_steps_to_load, list):
            ecco_dataset = open_mdsdataset(data_dir = mds_var_dir,
                                           grid_dir = mds_grid_dir,
                                           read_grid = read_grid,
                                           prefix = mds_files,
                                           geometry = 'llc',
                                           iters = model_time_steps_to_load,
                                           ref_date = ref_date,
                                           delta_t = delta_t,
                                           default_dtype = np.dtype(mds_datatype),
                                           grid_vars_to_coords = grid_vars_to_coords,
                                           llc_method=llc_method,
                                           ignore_unknown_vars=False,
                                           **kwargs)
        else:
            print('model time steps to load ', model_time_steps_to_load)
            raise TypeError('not a valid model_time_steps_to_load.  must be "all", an "int", or a list of "int"')

    if isinstance(cell_bounds, xr.core.dataset.Dataset):
        if 'XC_bnds' in cell_bounds.data_vars:
            ecco_dataset = ecco_dataset.assign_coords({"XC_bnds": (("tile","j","i","nb"), cell_bounds['XC_bnds'])})
        if 'YC_bnds' in cell_bounds.data_vars:
            ecco_dataset = ecco_dataset.assign_coords({"YC_bnds": (("tile","j","i","nb"), cell_bounds['YC_bnds'])})

    # replace the xmitgcm coordinate name of 'FACE' with 'TILE'
    if 'face' in ecco_dataset.coords.keys():
        ecco_dataset = ecco_dataset.rename({'face': 'tile'})
        ecco_dataset.tile.attrs['long_name'] = 'index of llc grid tile'



    if grid_vars_to_coords == False:
        # A bunch of grid geometry fields that aren't actually coordinates in
        # any sense are loaded by default as coordinates. We'll kick them all
        # down to data variables and then promote a few that are actually
        # coordinates.

        # Promote some variables as CF convention compliant coordinates.
        CF_legal_coords = ['XC','YC','XG','YG','Z','Zp1','Zu','Zl','time', 'iter']
        for legal_coord in CF_legal_coords:
            if legal_coord in list(ecco_dataset.data_vars):
                ecco_dataset = ecco_dataset.set_coords(legal_coord)

    # rename iter to timestep
    if 'iter' in ecco_dataset.coords.keys():
        ecco_dataset = ecco_dataset.rename({'iter': 'timestep'})



    # if vars_to_load is an empty list, keep all variables.  otherwise,
    # only keep those variables in the vars_to_load list.
    vars_ignored = []
    vars_loaded = []

    if not isinstance(vars_to_load, list):
        vars_to_load = [vars_to_load]

    if not less_output:
        print ('vars to load ', vars_to_load)

    if 'all' not in vars_to_load:
        if not less_output:
            print ('loading subset of variables: ', vars_to_load)

        # remove variables that are not on the vars_to_load_list
        for ecco_var in ecco_dataset.keys():

            if ecco_var not in vars_to_load:
                vars_ignored.append(ecco_var)
                ecco_dataset = ecco_dataset.drop_vars(ecco_var)

            else:
                vars_loaded.append(ecco_var)

        if not less_output:
            print ('loaded  : ', vars_loaded)
            print ('ignored : ', vars_ignored)

    else:
        if not less_output:
            print ('loaded all variables  : ', ecco_dataset.keys())

    # keep tiles in the 'tiles_to_load' list.
    if not isinstance(tiles_to_load, list) and not isinstance(tiles_to_load,range):
        tiles_to_load = [tiles_to_load]

    if not less_output:
        print ('subsetting tiles to ', tiles_to_load)

    ecco_dataset = ecco_dataset.sel(tile = tiles_to_load)

    # add time bounds for datasets with time-averaged quantities
    if not less_output:
        print ('creating time bounds .... ')

    if 'AVG' in output_freq_code and \
        'time_bnds' not in ecco_dataset.keys():

        if not less_output:
            print ('avg in output freq code and time bounds not in ecco keys')

        time_bnds_ds, center_times = \
            make_time_bounds_and_center_times_from_ecco_dataset(ecco_dataset,\
                                                                output_freq_code)

        ecco_dataset = xr.merge((ecco_dataset, time_bnds_ds))

        ecco_dataset = ecco_dataset.set_coords('time_bnds')

        if not less_output:
            print ('time bounds -----')
            print (time_bnds_ds)
            print ('center times -----')
            print (center_times)
            print ('ecco dataset time values type', type(ecco_dataset.time.values))
            print ('ecco dataset time_bnds typ   ', type(ecco_dataset.time_bnds))
            print ('ecco dataset time_bnds       ', ecco_dataset.time_bnds)

        if isinstance(ecco_dataset.time.values, np.datetime64):
            if not less_output:
                print ('replacing time.values....')
            ecco_dataset['time'].values = center_times

        elif isinstance(center_times, np.datetime64):
            if not less_output:
                print ('replacing time.values....')
            center_times = np.array(center_times)
            ecco_dataset['time'].values[:] = center_times

        elif isinstance(ecco_dataset.time.values, np.ndarray) and \
              isinstance(center_times, np.ndarray):
            if not less_output:
                print ('replacing time.values....')
            ecco_dataset = ecco_dataset.assign_coords({'time': center_times})
            #ecco_dataset['time'] = center_times


    # Drop mask Ctrl fields
    if 'maskCtrlS' in list(ecco_dataset.data_vars):
        ecco_dataset=ecco_dataset.drop_vars('maskCtrlS')
    if 'maskCtrlW' in list(ecco_dataset.data_vars):
        ecco_dataset=ecco_dataset.drop_vars('maskCtrlW')
    if 'maskCtrlC' in list(ecco_dataset.data_vars):
        ecco_dataset=ecco_dataset.drop_vars('maskCtrlC')

    # determine all of the dimensions used by data variables
    all_var_dims = set([])
    for ecco_var in ecco_dataset.data_vars:
        all_var_dims = set.union(all_var_dims, set(ecco_dataset[ecco_var].dims))

    if not less_output:
        print('all_var_dims ', all_var_dims)


    # update global, coordinate, and variable metadata
    # ... first, drop whatever is in 'standard_name' because those fields
    #     are generally not CF compliant.
    for ecco_var in ecco_dataset.data_vars.keys():
        if 'standard_name' in ecco_dataset[ecco_var].attrs.keys():
            ecco_dataset[ecco_var].attrs.pop('standard_name')
    for ecco_var in ecco_dataset.coords.keys():
        if 'standard_name' in ecco_dataset[ecco_var].attrs.keys():
            ecco_dataset[ecco_var].attrs.pop('standard_name')

    # update coordinate metadata
    if len(coordinate_metadata) > 0:
        ecco_dataset = add_coordinate_metadata(coordinate_metadata, ecco_dataset)

    # update variable metadata
    if len(variable_metadata) > 0:
        ecco_dataset, gcmd_keywords = \
            add_variable_metadata(variable_metadata, ecco_dataset)

    # update global metadata
    # possible keys
    keys_2D = set(['i','j','i_g','j_g'])
    keys_3D = set(['k','k_l','k_u','k_p1'])

    if len(all_var_dims.intersection(keys_3D)) > 0:
        dataset_dim = '3D'
    elif len(all_var_dims.intersection(keys_2D)) > 0:
        dataset_dim = '2D'
    else:
        #print('no 2D or 3D dimensions')
        print('cannot find 2D or 3D dims in dataset')
        sys.exit()

    # drop 3D and dims coordinates that do not appear in any data variable
    if drop_unused_coords:

        if dataset_dim == '2D':
            if not less_output:
                print('\n only 2D variables, dropping 3D dims and coords')
                print(all_var_dims)

            # drop 3D dimensions
            dims_to_drop = set(ecco_dataset.dims).intersection(set(['k','k_u','k_l','k_p1']))
            if not less_output:
                print('\n dropping 3D dims')
                print(dims_to_drop)

            for dim in dims_to_drop:
                if not less_output:
                    print('--> dropping', dim)
                ecco_dataset = ecco_dataset.drop(dim)

            # drop 3D coords
            if not less_output:
                print('\n dropping 3D coords')

            coords_to_drop = []
            for coord in list(ecco_dataset.coords):
                coord_dims = set(ecco_dataset.coords[coord].dims)
                coord_tmp = coord_dims.intersection(dims_to_drop)
                if len(coord_tmp) > 0:
                    if not less_output:
                        print('--> dropping ', coord)
                    ecco_dataset = ecco_dataset.drop(coord)

    # apply global metadata
    if len(global_metadata) > 0:

        #print('dataset dim: ', dataset_dim)
        # some metadata is 3D dataset specific.
        ecco_dataset = \
            add_global_metadata(global_metadata, ecco_dataset, dataset_dim,\
                                less_output=less_output)

    if not less_output:
        print('dataset dim: ', dataset_dim)
        print('dataset_coords : ', ecco_dataset.coords)

    # add current time and date
    current_time = datetime.datetime.now().isoformat()[0:19]
    ecco_dataset.attrs['date_created'] = current_time
    ecco_dataset.attrs['date_modified'] = current_time
    ecco_dataset.attrs['date_metadata_modified'] = current_time
    ecco_dataset.attrs['date_issued'] = current_time

    if not less_output:
        print('output_freq_code: ', output_freq_code)
    # set the long name of the time attribute
    if 'AVG' in output_freq_code:
        ecco_dataset.time.attrs['long_name'] = 'center time of averaging period'
    elif 'SNAP' in output_freq_code:
        ecco_dataset.time.attrs['long_name'] = 'snapshot time'

    # set averaging period duration and resolution
    if not less_output:
        print('\n... setting time coverage resolution')
    # --- AVG DAY
    if output_freq_code == 'AVG_MON':
        ecco_dataset.attrs['time_coverage_duration'] = 'P1M'
        ecco_dataset.attrs['time_coverage_resolution'] = 'P1M'
    # --- AVG DAY
    elif output_freq_code == 'AVG_DAY':
        ecco_dataset.attrs['time_coverage_duration'] = 'P1D'
        ecco_dataset.attrs['time_coverage_resolution'] = 'P1D'
    # --- SNAPSHOT
    elif output_freq_code == 'SNAPSHOT':
        ecco_dataset.attrs['time_coverage_duration'] = 'P0S'
        ecco_dataset.attrs['time_coverage_resolution'] = 'P0S'

    ecco_dataset.attrs['original_mds_var_dir'] = str(mds_var_dir)
    ecco_dataset.attrs['original_mds_grid_dir'] = str(mds_grid_dir)

    # alaphbetically sort global attributes
    ecco_dataset.attrs = sort_attrs(ecco_dataset.attrs)


    if not less_output:
        for dim in ecco_dataset.dims:
            print(dim, ecco_dataset[dim].attrs)

    return ecco_dataset

#%%
def read_llc_to_compact(fdir, fname, llc=90, skip=0, nk=1, nl=1,
            filetype = '>f4', less_output = False ):
    """

    Loads an MITgcm binary file in the compact format of the
    lat-lon-cap (LLC) grids.  Note, does not use Dask.

    Array is returned with the following dimension order:

        [nl, nk, N_tiles*llc, llc]

    where if either N_z or N_recs =1, then that dimension is collapsed
    and not present in the returned array.

    Parameters
    ----------
    fdir : string
        A string with the directory of the binary file to open

    fname : string
        A string with the name of the binary file to open

    llc : int, optional, default 90
        the size of the llc grid.  For ECCO v4, we use the llc90 domain
        so `llc` would be `90`.
        Default: 90

    skip : int, optional, default 0
        the number of 2D slices (or records) to skip.  Records could be vertical levels of a 3D field, or different 2D fields, or both.

    nk : int, optional, default 1 [singleton]
        number of 2D slices (or records) to load in the depth dimension.
        Default: 1 [singleton]

    nl : int, optional, default 1 [singleton]
        number of 2D slices (or records) to load in the "record" dimension.
        Default: 1 [singleton]

    filetype: string, default '>f4'
        the file type, default is big endian (>) 32 bit float (f4)
        alternatively, ('<d') would be little endian (<) 64 bit float (d)

    less_output : boolean, optional, default False
        A debugging flag.  False = less debugging output


    Returns
    -------
    data_compact : ndarray
        a numpy array of dimension nl x nk x 13*llc x llc

    """
    data_compact = load_binary_array(fdir, fname, llc, 13*llc, nk=nk, nl=nl,
                                     skip=skip, filetype = filetype,
                                     less_output = less_output)
    #data_tiles = read_bin_to_tiles(fdir,fname,llc=llc,nk=nk,nl=nl,skip=skip,
    #                skip=skip, filetype = filetype, less_output = less_output)

    #data_tiles = read_llc_to_tiles(fdir,fname,llc=llc,nk=nk,nl=nl,skip=skip,
    #                               filetype=filetype)

    #data_compact = llc_tiles_to_compact(data_tiles,less_output=less_output)

    # return the array
    return data_compact


#%%
def read_llc_to_tiles(fdir, fname, llc=90, skip=0, nk=1, nl=1,
              	      filetype = '>f', less_output = False,
                      use_xmitgcm=False):
    """


    Loads an MITgcm binary file in the 'tiled' format of the
    lat-lon-cap (LLC) grids with dimension order:

        [N_recs, N_z, N_tiles, llc, llc]

    where if either N_z or N_recs =1, then that dimension is collapsed
    and not present in the returned array.

    if use_xmitgcm == True

        data are read in via the low level routine
        xmitgcm.utils.read_3d_llc_data and returned as dask array.

        Hint: use data_tiles.compute() to load into memory.

    if use_xmitgcm == False

        Loads an MITgcm binary file in the 'compact' format of the
        lat-lon-cap (LLC) grids and converts it to the '13 tiles' format
        of the LLC grids.

    Parameters
    ----------
    fdir : string
        A string with the directory of the binary file to open
    fname : string
        A string with the name of the binary file to open
    llc : int
        the size of the llc grid.  For ECCO v4, we use the llc90 domain
        so `llc` would be `90`.
        Default: 90
    skip : int
        the number of 2D slices (or records) to skip.
        Records could be vertical levels of a 3D field, or different 2D fields, or both.
    nk : int
        number of 2D slices (or records) to load in the third dimension.
        if nk = -1, load all 2D slices
        Default: 1 [singleton]
    nl : int
        number of 2D slices (or records) to load in the fourth dimension.
        Default: 1 [singleton]
    filetype: string
        the file type, default is big endian (>) 32 bit float (f)
        alternatively, ('<d') would be little endian (<) 64 bit float (d)
    less_output : boolean
        A debugging flag.  False = less debugging output
        Default: False
    use_xmitgcm : boolean
        option to use the routine xmitgcm.utils.read_3d_llc_data into a dask
        array, i.e. not into memory.
        Otherwise read in as a compact array, convert to faces, then to tiled format
        Default: False

    Returns
    -------
    data_tiles
        a numpy array of dimension 13 x nl x nk x llc x llc, one llc x llc array
        for each of the 13 tiles and nl and nk levels.
    """

    if use_xmitgcm:

        full_filename = '%s/%s' % (fdir,fname)

        if not less_output:
            print('read_llc_to_tiles: full_filename: ',full_filename)

        # Handle "skipped" records by reading up until that record, and
        # dropping preceding records afterward
        #
        # Note: that xmitgcm looks at recs including a full 3D chunk
        # while "skip" refers to 2D chunks.
        nrecs = nl
        skip_3d = int(np.ceil(skip/nk))
        nrecs += skip_3d

        # Reads data into dask array as numpy memmap
        # [Nrecs x Nz x Ntiles x llc x llc]
        data_tiles = xmitgcm.utils.read_3d_llc_data(full_filename, nx=llc, nz=nk,
                                                    nrecs=nrecs, dtype=filetype,
                                                    memmap=False)

        # Handle cases of single or multiple records, and skip>0
        if skip>0:
            if nk>1 and nl>1:
                data_tiles = np.reshape(data_tiles,(nk*nrecs,)+data_tiles.shape[-3:])
                data_tiles = data_tiles[skip:skip+nl*nk]
                data_tiles = np.reshape(data_tiles,(nl,nk)+data_tiles.shape[-3:])
            else:
                data_tiles = data_tiles[skip:skip+nl,...]

    else:

        data_compact = read_llc_to_compact(fdir, fname, llc=llc, skip=skip, nk=nk, nl=nl,
           				    filetype = filetype, less_output=less_output)

        data_tiles   = llc_compact_to_tiles(data_compact, less_output=less_output)

    # return the array
    return data_tiles

#%%
def read_llc_to_faces(fdir, fname, llc=90, skip=0, nk=1, nl=1,
        filetype = '>f4', less_output = False):
    """

    Loads an MITgcm binary file in the 'compact' format of the
    lat-lon-cap (LLC) grids and converts it to the '5 faces' format
    of the LLC grids.

    Can load 2D and 3D arrays.

    Array is returned with the following dimension order:
    - [N_faces][N_recs, N_z, N_y, N_x]

    where if either N_z or N_recs =1, then that dimension is collapsed
    and not present in the returned array.

    Parameters
    ----------
    fdir : string
        A string with the directory of the binary file to open

    fname : string
        A string with the name of the binary file to open

    llc : int, optional, default 90
        the size of the llc grid.  For ECCO v4, we use the llc90 domain (llc=90)

    skip : int, optional, default 0
        the number of 2D slices (or records) to skip.
        Records could be vertical levels of a 3D field, or different 2D fields, or both.

    nk : int, optional, default 1 [singleton]
        number of 2D slices (or records) to load in the depth dimension.

    nl : int, optional, default 1 [singleton]
        number of 2D slices (or records) to load in the "record" dimension.

    filetype : string, default '>f4'
        the file type, default is big endian (>) 32 bit float (f4)
        alternatively, ('<d') would be little endian (<) 64 bit float (d)

    less_output : boolean, optional, default False
        A debugging flag.  False = less debugging output


    Returns
    -------
    data_faces : dict
        a dictionary containing the five lat-lon-cap faces data_faces[n]
        dimensions of each 2D slice of data_faces
        f1,f2: 3*llc x llc
        f3: llc x llc
        4,f5: llc x 3*llc

    """

    data_compact = read_llc_to_compact(fdir, fname, llc=llc, skip=skip, nk=nk, nl=nl,
                                    filetype = filetype, less_output=less_output)

    data_faces = llc_compact_to_faces(data_compact, less_output = less_output)

    #data_tiles = read_llc_to_tiles(fdir,fname,llc=llc,nk=nk,nl=nl,skip=skip,
    #                               filetype=filetype)

    #data_faces = llc_tiles_to_faces(data_tiles, less_output = less_output)

    return data_faces



import os
import shutil
import numpy as np
import ecco_v4_py
import pytest

from .test_common import (
        llc_mds_datadirs, get_global_ds,
        all_mds_datadirs, get_test_ds)

_test_dir = os.path.dirname(os.path.abspath(__file__))

# test error out for different domains
def test_notimplemented(get_test_ds):
    ds = get_test_ds
    with pytest.raises(NotImplementedError):
        if len(ds.tile)<13:
            ecco_v4_py.get_basin_mask('atl',ds.maskC)
        else:
            ecco_v4_py.get_basin_mask('atl',ds.sel(tile=0).maskC)
            ecco_v4_py.get_basin_mask('atl',(1.*ds.maskC).diff(dim='i'))
            ecco_v4_py.get_basin_mask('atl',(1.*ds.maskC).diff(dim='j'))

def test_each_basin_masks(get_global_ds):
    """make sure we can make the basin masks
    """

    ds = get_global_ds
    all_basins = ecco_v4_py.read_llc_to_tiles(os.path.join(_test_dir,'..','..','binary_data'),'basins.data',less_output=True)
    ext_names = ['atlExt','pacExt','indExt']
    for i,basin in enumerate(ecco_v4_py.get_available_basin_names(),start=1):
        mask = ecco_v4_py.get_basin_mask(basin,ds.maskC.isel(k=0))
        assert np.all(mask.values == (all_basins==i))

def test_ext_basin_masks(get_global_ds):
    """make sure we can make the extended masks
    """

    ds = get_global_ds

    ext_names = ['atlExt','pacExt','indExt']
    individual_names = [['atl','mexico','hudson','med','north','baffin','gin'],
                        ['pac','bering','okhotsk','japan','eastChina'],
                        ['ind','southChina','java','timor','red','gulf']]
    for ext,ind in zip(ext_names,individual_names):
        maskE = ecco_v4_py.get_basin_mask(ext,ds.maskC.isel(k=0))
        maskI = ecco_v4_py.get_basin_mask(ind,ds.maskC.isel(k=0))
        assert np.all(maskE==maskI)

def test_3d(get_global_ds):
    """check that vertical coordinate"""

    ds = get_global_ds
    grid = ecco_v4_py.get_llc_grid(ds)
    maskK = ds['maskC']
    maskL = grid.interp(maskK,'Z',to='left',boundary='fill')
    maskU = grid.interp(maskK,'Z',to='right',boundary='fill')
    maskKp1 = ds.maskC.isel(k=0).broadcast_like(ds.k_p1)
    for mask in [maskK,maskL,maskU,maskKp1]:
        ecco_v4_py.get_basin_mask('atl',mask)

def test_bin_dir_is_here(get_global_ds,hide_bin_dir):

    hide_bin_dir
    ds = get_global_ds
    with pytest.raises(OSError):
        ecco_v4_py.get_basin_mask('atl',ds.maskC.isel(k=0))



@pytest.fixture(scope='function')
def hide_bin_dir():

    bin_path = os.path.join(_test_dir,'..','..','binary_data')
    tmp_path = os.path.join(_test_dir,'..','..','tmp_bin')
    shutil.move(bin_path,tmp_path)

    yield

    shutil.move(tmp_path,bin_path)



### This module contains routines to search the text of ECCO variable lists for matching datasets


import requests
import sys
from os.path import join


def varlist_file_parse(file):
    """Parse content of varlist file into dictionary"""

    response = requests.get(file)
    varlist_lines = response.text.split('\n')
    content_dict = {}
    curr_shortname = ''
    multiline_var = False
    for line in varlist_lines:
        if len(line) == 0:
            continue
        if (('ShortName' in line) and ('Variable Name' in line) and ('Description' in line)):
            varname_pos = line.index('Variable Name')
            descrip_pos = line.index('Description')
            varname_colwidth = descrip_pos - varname_pos
        elif 'ECCO_' in line:
            if '*' in line:
                first_ast_pos = line.index('*')
                curr_shortname = line[:first_ast_pos].replace(' ','')
                content_dict[curr_shortname] = {'Note':line[first_ast_pos:]}
            else:
                curr_shortname = line
                content_dict[curr_shortname] = {}
        elif 'Note:' in line:
            content_dict[curr_shortname]['Note'] = line[6:]
        elif len(curr_shortname) > 0:
            if len(line[:varname_pos].replace(' ','')) > 0:
                continue
            len_varname = line[varname_pos:].find(' ')
            if len_varname == -1:
                len_varname = len(line[varname_pos:])
            if len_varname >= varname_colwidth:
                varname = line[varname_pos:]
                multiline_var = True
                continue
            elif not multiline_var:
                varname = line[varname_pos:descrip_pos].replace(' ','')
            descrip = line[descrip_pos:]
            content_dict[curr_shortname][varname] = descrip
            if multiline_var:
                multiline_var = False

    return content_dict,varname_pos,descrip_pos



###================================================================================================================


def query_shortnames(query,content_dict,\
                     grid,test_grids,\
                     time_res,test_timeres):
    """Search varlist file for case-insensitive query results"""
    
    # make query case-insensitive
    query_ci = query.casefold()
    
    exactmatch_flag = False
    shortnames_match = []
    for shortname,curr_content in content_dict.items():

        # handle query results from the "mixed" grid and time res files
        # if single grid or time res was specified
        if ((grid is not None) and ('mixed' in test_grids)):
            if 'Note' in curr_content.keys():
                if '*' in curr_content['Note']:
                    if ((grid == 'native')\
                      and ('native' not in curr_content['Note'])):
                        continue
                    elif ((grid == 'latlon')\
                      and ('lat-lon' not in curr_content['Note'])):
                        continue
        if ((time_res != 'all') and ('all' in test_timeres)):
            if 'Note' in curr_content.keys():
                if '*' in curr_content['Note']:
                    if time_res not in curr_content['Note']:
                        continue

        # add query result to list
        if query_ci == shortname.casefold():
            exactmatch_flag = True
            shortnames_match.append(shortname)
        else:
            if query_ci in shortname.casefold():
                shortnames_match.append(shortname)
            elif query_ci in str(curr_content).casefold():
                shortnames_match.append(shortname)
    
    return shortnames_match,exactmatch_flag



###================================================================================================================


def print_varlist_query_results(query,shortnames_match,\
                                content_dict,varname_pos,descrip_pos):
    """Print text output of query"""

    varname_colwidth = descrip_pos - varname_pos
    query_resp_str = 'ShortName Options for query "'+query+'":\n'\
                        +(' '*varname_pos)\
                        +'Variable Name'.ljust(descrip_pos-varname_pos)\
                        +'Description (units)\n\n'
    for count,shortname in enumerate(shortnames_match):
        shortname_line = 'Option '+str(count+1)+': '+shortname
        if (('Note' in content_dict[shortname].keys())\
          and (' *' in content_dict[shortname]['Note'])):
            shortname_line += ('    '+content_dict[shortname]['Note'])
        else:
            shortname_line += '    *'
            if 'LLC' in shortname:
                shortname_line += 'native grid,'
            elif 'DEG' in shortname:
                shortname_line += 'lat-lon grid,'
            if 'MONTHLY' in shortname:
                shortname_line += 'monthly means*'
            elif 'DAILY' in shortname:
                shortname_line += 'daily means*'
            elif 'SNAPSHOT' in shortname:
                shortname_line += 'snapshots at daily intervals*'
        shortname_line += '\n'
        query_resp_str += shortname_line
        for varname,descrip in content_dict[shortname].items():
            max_descrip_len_perline = 50
            if varname == 'Note':
                max_descrip_len_perline += descrip_pos - 6
            if len(descrip) > max_descrip_len_perline:
                # put line breaks in long descriptions
                descrip_withbreaks = ''
                descrip_remaining = descrip + ''
                while len(descrip_remaining) > 0:
                    descrip_words = descrip_remaining[:(max_descrip_len_perline+1)].split(' ')
                    if len(descrip_remaining) <= max_descrip_len_perline:
                        curr_descrip = descrip_remaining
                    else:
                        curr_descrip = " ".join(descrip_words[:-1])
                    descrip_withbreaks += curr_descrip
                    descrip_remaining = descrip_remaining[(len(curr_descrip)+1):]
                    if len(descrip_remaining) > 0:
                        if varname == 'Note':
                            descrip_withbreaks += '\n      '
                        else:
                            descrip_withbreaks += ('\n'+(' '*descrip_pos))
                descrip = descrip_withbreaks
            if varname == 'Note':
                if ' *' in descrip:
                    continue
                else:
                    var_line = 'Note: '+descrip+'\n'
            elif len(varname) >= varname_colwidth:
                var_line = (' '*varname_pos)+varname+'\n'
                var_line += (' '*descrip_pos)+descrip+'\n'
            else:
                var_line = (' '*varname_pos)+(varname.ljust(varname_colwidth,' '))\
                            +descrip+'\n'
            query_resp_str += var_line
        query_resp_str += '\n'

    print(query_resp_str)
    
    return None



###================================================================================================================


def ecco_podaac_varlist_query(query,version,grid=None,time_res='all'):
    """
    
    This function queries the ECCO variable lists to find matches to "query" that match the grid and time_res types.
    The function then takes input from the user to determine which matching result to use.
    If no matches are found, then a ValueError raised.
    
    Parameters
    ----------    
    query: str, a text string being used to query ShortNames, variable names,
           and descriptions.
    
    version: ('v4r4','v4r5'), ECCO version to search variable lists for
    
    grid: ('native','latlon',None), specifies whether to query datasets with output
          on the native grid or the interpolated lat/lon grid.
          The default None will query both types of grids (and datasets with no spatial
          dimension), unless specified otherwise in a query dict (e.g., the example above).
    
    time_res: ('monthly','daily','snapshot','all'), specifies which time resolution 
              to include in query and downloads. 'all' includes all time resolutions, 
              and datasets that have no time dimension, such as the grid parameter 
              and mixing coefficient datasets.
    
    Returns
    -------
    shortname_match: str, ShortName matching the ECCO variable list query.
    
    """
    
    
    pass
    
    
    if version not in ['v4r4','v4r5']:
        raise ValueError('ECCO '+version+' is not currently available from PO.DAAC')
    
    # paths to variable list files
    varlist_url_root = 'https://raw.githubusercontent.com/ECCO-GROUP/ECCOv4-py/refs/heads/master/ecco_v4_py/varlist/'
    if version == 'v4r4':
        varlist_url_ids = {'native,monthly':'v4r4/v4r4_nctiles_monthly_varlist.txt',\
                           'native,daily':'v4r4/v4r4_nctiles_daily_varlist.txt',\
                           'native,snapshot':'v4r4/v4r4_nctiles_snapshots_varlist.txt',\
                           'latlon,monthly':'v4r4/v4r4_latlon_monthly_varlist.txt',\
                           'latlon,daily':'v4r4/v4r4_latlon_daily_varlist.txt',\
                           'mixed,all':'v4r4/v4r4_tseries_grid_varlist.txt'}
    elif version == 'v4r5':
        varlist_url_ids = {'native,monthly':'v4r5/v4r5_nctiles_monthly_varlist.txt',\
                           'native,daily':'v4r5/v4r5_nctiles_daily_varlist.txt',\
                           'native,snapshot':'v4r5/v4r5_nctiles_snapshots_varlist.txt',\
                           'latlon,monthly':'v4r5/v4r5_latlon_monthly_varlist.txt',\
                           'latlon,daily':'v4r5/v4r5_latlon_daily_varlist.txt'}
    
    
    # set keys of grid types and time resolutions to search
    
    if grid is None:
        test_grids = ['native','latlon','mixed']
        if time_res != 'all':
            test_timeres = [time_res,'all']
    else:
        test_grids = [grid]
    if time_res == 'all':
        if grid is not None:
            test_grids += ['mixed']
        test_timeres = ['monthly','daily','snapshot','all']
    elif grid is not None:
        test_timeres = [time_res]
    
    grid_timeres_keys = []
    for gridtype in test_grids:
        for timeres in test_timeres:
            if (((gridtype == 'mixed') and (timeres != 'all'))\
              or ((gridtype != 'mixed') and (timeres == 'all'))):
                # include time series/grid varlist
                grid_timeres_keys.append('mixed,all')
            elif ((gridtype == 'latlon') and (timeres == 'snapshot')):
                continue        
            else:
                grid_timeres_keys.append(gridtype+','+timeres)    
    
    # remove any duplicate keys
    keys_no_duplicates = []
    keys_no_duplicates = [key for key in grid_timeres_keys\
                              if key not in keys_no_duplicates]
    grid_timeres_keys = keys_no_duplicates
    
    # build content dictionary containing all varlists being queried
    content_dict = {}
    for curr_key in grid_timeres_keys:
        if curr_key not in varlist_url_ids.keys():
            continue
        curr_content_dict,varname_pos,descrip_pos = varlist_file_parse(\
                                    varlist_url_root+varlist_url_ids[curr_key])
        content_dict = {**content_dict,**curr_content_dict}
    
    # find matches to query in the varlists
    shortnames_match,exactmatch_flag = query_shortnames(query,content_dict,\
                                                        grid,test_grids,\
                                                        time_res,test_timeres)
    
    if exactmatch_flag:
        shortname_match = shortnames_match[0]
    else:
        # print to screen the query match results
        print_varlist_query_results(query,shortnames_match,\
                                    content_dict,varname_pos,descrip_pos)
        
        if len(shortnames_match) == 0:
            raise ValueError('No valid matches to query found; please try again.')
        if len(shortnames_match) == 1:
            option_proceed = input('Proceed with option 1? [y/n]: ')
            if option_proceed.casefold() == 'y':
                option_num = 1
            else:
                raise ValueError('No valid matches to query found; please try again.')
        else:
            option_num = input('Please select option [1-'+str(len(shortnames_match))+']: ')
        shortname_match = shortnames_match[int(option_num)-1]
        print('Using dataset with ShortName: '+shortname_match)
    
    return shortname_match



from .calc_meridional_trsp import calc_meridional_vol_trsp, \
        calc_meridional_heat_trsp, calc_meridional_salt_trsp, \
        meridional_trsp_at_depth

from .calc_section_trsp import calc_section_vol_trsp, \
        calc_section_heat_trsp, calc_section_salt_trsp, \
        section_trsp_at_depth

from .calc_stf import calc_meridional_stf, calc_section_stf

from .ecco_utils import make_time_bounds_and_center_times_from_ecco_dataset
from .ecco_utils import make_time_bounds_from_ds64
from .ecco_utils import extract_yyyy_mm_dd_hh_mm_ss_from_datetime64

from .ecco_utils import minimal_metadata
from .ecco_utils import months2days
from .ecco_utils import get_llc_grid
from .ecco_utils import find_metadata_in_json_dictionary
from .ecco_utils import add_global_metadata
from .ecco_utils import add_coordinate_metadata
from .ecco_utils import add_variable_metadata
from .ecco_utils import sort_attrs
from .ecco_utils import sort_all_attrs

from .get_basin import get_basin_mask, get_available_basin_names

from .get_section_masks import get_section_endpoints, \
        get_available_sections, get_section_line_masks

#from extract_grid_fields_from_mitgrid_files import extract_U_point_grid_fields_from_mitgrid_as_tiles
#from extract_grid_fields_from_mitgrid_files import extract_G_point_grid_fields_from_mitgrid_as_tiles

from .llc_array_conversion  import llc_compact_to_tiles
from .llc_array_conversion  import llc_tiles_to_compact
from .llc_array_conversion  import llc_compact_to_faces
from .llc_array_conversion  import llc_faces_to_tiles
from .llc_array_conversion  import llc_ig_jg_faces_to_tiles
from .llc_array_conversion  import llc_tiles_to_faces
from .llc_array_conversion  import llc_faces_to_compact
from .llc_array_conversion  import llc_tiles_to_xda

#from .netcdf_product_generation import create_nc_grid_files_on_native_grid_from_mds
from .netcdf_product_generation import get_time_steps_from_mds_files
#from .netcdf_product_generation import create_nc_variable_files_on_native_grid_from_mds
from .netcdf_product_generation import create_nc_variable_files_on_regular_grid_from_mds
#from .netcdf_product_generation import update_ecco_dataset_geospatial_metadata
#from .netcdf_product_generation import update_ecco_dataset_temporal_coverage_metadata
from .netcdf_product_generation import save_ecco_dataset_to_netcdf
from .netcdf_product_generation import make_date_str_from_dt64

from .read_bin_llc import read_llc_to_tiles
from .read_bin_llc import read_llc_to_compact
from .read_bin_llc import read_llc_to_faces
from .read_bin_llc import load_ecco_vars_from_mds
from .read_bin_gen import load_binary_array

from .resample_to_latlon import resample_to_latlon

from ecco_v4_py import  scalar_calc

from .tile_io import load_ecco_grid_nc
from .tile_io import load_ecco_var_from_years_nc
from .tile_io import recursive_load_ecco_var_from_years_nc

from .tile_plot import plot_tile
from .tile_plot import plot_tiles
from .tile_plot import unique_color

from .tile_plot_proj import plot_proj_to_latlon_grid
from .tile_plot_proj import plot_pstereo
from .tile_plot_proj import plot_global

from ecco_v4_py import vector_calc

from .ecco_access import ecco_podaac_access
from .ecco_access import ecco_podaac_to_xrdataset

__all__ = ['calc_meridional_trsp',
           'calc_section_trsp',
           'calc_stf',
           'ecco_acc_dates',
           'ecco_access',
           'ecco_download',
           'ecco_s3_retrieve',
           'ecco_utils',
           'ecco_varlist',
           'get_section_masks',
           'llc_array_conversion',
           'netcdf_product_generation',
           'read_bin_llc',
           'read_bin_gen',
           'resample_to_latlon',
           'scalar_calc',
           'tile_io',
           'tile_plot',
           'tile_plot_proj',
           'vector_calc']



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""ECCO v4 Python: Dataset Utililites

This module includes utility routines to create the ECCO product as netcdf and
for processing metadata.

.. _ecco_v4_py Documentation :
   https://github.com/ECCO-GROUP/ECCOv4-py
"""

from __future__ import division, print_function
import numpy as np
import xarray as xr
import datetime
import time
import xmitgcm as xm
import dateutil
import glob
import os
import sys
import pyresample as pr
import json
from pathlib import Path
import uuid
import netCDF4 as nc4
from collections import OrderedDict
import dask as dask
from dask import delayed

from .read_bin_llc import load_ecco_vars_from_mds
from .ecco_utils import extract_yyyy_mm_dd_hh_mm_ss_from_datetime64
from .resample_to_latlon import resample_to_latlon

#%%
def sort_attrs(attrs):
    od = OrderedDict()

    keys = sorted(list(attrs.keys()),key=str.casefold)

    for k in keys:
        od[k] = attrs[k]

    return od

#%%


def create_native_grid_netcdf_files(mds_grid_dir, mds_var_dir, mds_filename,
                                    mds_freq_code,
                                    vars_to_load,
                                    dataset_name = 'by_variable',
                                    time_steps_to_load = 'all',
                                    tiles_to_load = 'all',
                                    output_array_precision = np.float32,
                                    global_metadata = 'default',
                                    coordinate_metadata = 'default',
                                    geometry_metadata = 'default',
                                    variable_metadata = 'default'):


    # if no specific file data passed, read default metadata from json file
    # -- variable specific meta data
    script_dir = Path(__file__).resolve().parent

    #if not meta_variable_specific:
    #    meta_variable_rel_path = '../meta_json/ecco_meta_variable.json'
    #    abs_meta_variable_path = os.path.join(script_dir, meta_variable_rel_path)
    #    with open(abs_meta_variable_path, 'r') as fp:
    #        meta_variable_specific = json.load(fp)


    ## METADATA
    metadata_json_dir = Path('/home/ifenty/git_repos_others/ECCO-GROUP/ECCO-ACCESS/metadata/ECCOv4r4_metadata_json')

    metadata_fields = ['ECCOv4_global_metadata_for_all_datasets',
                       'ECCOv4_global_metadata_for_latlon_datasets',
                       'ECCOv4_global_metadata_for_native_datasets',
                       'ECCOv4rcoordinate_metadata_for_native_datasets',
                       'ECCOv4r4_geometry_metadata_for_native_datasets',
                       'ECCOv4r4_variable_metadata']

    print('\nLOADING METADATA')
    # load METADATA
    metadata = dict()

    for mf in metadata_fields:
        mf_e = mf + '.json'
        print(mf_e)
        with open(str(metadata_json_dir / mf_e), 'r') as fp:
            metadata[mf] = json.load(fp)


    # metadata for different variables
    global_metadata_for_all_datasets = metadata['ECCOv4r4_global_metadata_for_all_datasets']
    global_metadata_for_latlon_datasets = metadata['ECCOv4r4_global_metadata_for_latlon_datasets']
    global_metadata_for_native_datasets = metadata['ECCOv4r4_global_metadata_for_native_datasets']

    coordinate_metadata_for_1D_datasets = metadata['ECCOv4r4_coordinate_metadata_for_1D_datasets']
    coordinate_metadata_for_latlon_datasets = metadata['ECCOv4r4_coordinate_metadata_for_latlon_datasets']
    coordinate_metadata_for_native_datasets = metadata['ECCOv4r4_coordinate_metadata_for_native_datasets']

    geometry_metadata_for_latlon_datasets = metadata['ECCOv4r4_geometry_metadata_for_latlon_datasets']
    geometry_metadata_for_native_datasets = metadata['ECCOv4r4_geometry_metadata_for_native_datasets']

    groupings_for_1D_datasets = metadata['ECCOv4r4_groupings_for_1D_datasets']
    groupings_for_latlon_datasets = metadata['ECCOv4r4_groupings_for_latlon_datasets']
    groupings_for_native_datasets = metadata['ECCOv4r4_groupings_for_native_datasets']

    variable_metadata_latlon = metadata['ECCOv4r4_variable_metadata_for_latlon_datasets']
    variable_metadata = metadata['ECCOv4r4_variable_metadata']

    global_metadata = global_metadata_for_all_datasets + global_metadata_for_native_datasets

    variable_metadata_combined = variable_metadata + geometry_metadata_for_native_datasets


    #  #variable_metadata = variable_metadata + geometry_metadata_for_native_datasets
    short_mds_name = 'ETAN_mon_mean'
    output_freq_code= 'AVG_MON'
    cur_ts = 'all'
    vars_to_load = 'ETAN' #['ETAN']
    F_DS = []

    ecco_grid =  load_ecco_vars_from_mds(str(mds_grid_dir),
                                    str(mds_grid_dir),
                                    '',
                                    vars_to_load = 'all',
                                    drop_unused_coords = False,
                                    grid_vars_to_coords = False,
                                    coordinate_metadata = coordinate_metadata_for_native_datasets,
                                    variable_metadata = geometry_metadata_for_native_datasets,
                                    global_metadata = global_metadata,
                                    less_output=False).load()


    F_DS = \
      load_ecco_vars_from_mds(mds_var_dir,\
                                   mds_grid_dir = mds_grid_dir, \
                                   mds_files = short_mds_name,\
                                   vars_to_load = vars_to_load,
                                   drop_unused_coords = True,\
                                   grid_vars_to_coords = False,\
                                   variable_metadata = variable_metadata_combined,
                                   coordinate_metadata = coordinate_metadata_for_native_datasets,
                                   #global_metadata = [[global_metadata]],
                                   output_freq_code=output_freq_code,\
                                   model_time_steps_to_load=cur_ts,
                                   less_output = True)

    print(F_DS)
    #  vars_to_drop = set(F_DS.data_vars).difference(set([var]))
    #  F_DS.drop_vars(vars_to_drop)


    save_ecco_dataset_to_netcdf(F_DS.isel(time=[0]),
                                    Path('/home/ifenty/tmp/'),
                                    dataset_name = 'by_variable',
                                    time_method = 'by_year',
                                    output_freq_code='AVG_MON')




#%%%%%%%%%%%%%%%%%%%%
def save_ecco_dataset_to_netcdf(ecco_ds,
                                output_dir,
                                dataset_name = 'by_variable',
                                time_method = 'by_record',
                                output_array_precision = np.float32,
                                output_freq_code=None):

    """Saves an ECCO dataset to one or more NetCDF files

    NetCDF files will be written with the following options
    -------------------------------------------------------
    * compression level 5
    * shuffle = True
    * zlib = True

    Parameters
    ----------
    ecco_ds: xarray DataSet
        the DataSet to save.  Can have one or more 'data variables'

    output_dir: String
        root directory for saved files.  New files will be saved in
        a (new) subdirectory of output_dir

    dataset_name : String, optional.  Default 'by_variable'
        name to use for NetCDF files.  'by_variable' will create a
        name based on the data variables present by concatenating
        all data variable names together separated by '_'
        For example, if ecco_ds has both 'ETAN' and 'SSH' the
        dataset_name will be 'ETAN_SSH'

	time_method : String, optional. Default 'by_record'
        options include
            'by_record' - one file per time level
            'by_year'   - one file per calendar year

    output_array_precision : numpy type. Default np.float32
        precision to use when saving data variables of type float
        options include
            np.float32
            np.float64

    output_freq_code: String, optional. Default = None
        a string code indicating the time level of averaging of the
        data variables
        options include
            'AVG_MON'  - monthly-averaged file
            'AVG_DAY'  - daily-averaged files
            'SNAP'     - snapshot files (instantaneous)

    RETURNS:
    ----------
    nothing.  files should be saved to disk

    """


    # Create a name of the files if not specified
    # ---------------------------------------------
    if dataset_name =='by_variable':
        # concat all data variables together into a single string
        dataset_name = '_'.join(list(ecco_ds.data_vars))


    # force load coordinate values in case they are in dask array
    # -----------------------------------------------------------
    for coord in ecco_ds.coords:
        ecco_ds[coord].load()


    # Define fill values for NaN
    # ---------------------------------------------
    if output_array_precision == np.float32:
        netcdf_fill_value = nc4.default_fillvals['f4']

    elif output_array_precision == np.float64:
        netcdf_fill_value = nc4.default_fillvals['f8']


    # Create NetCDF encoding directives
    # ---------------------------------------------
    print('\n... creating variable encodings')
    # ... data variable encoding directives
    dv_encoding = dict()
    for dv in ecco_ds.data_vars:
        dv_encoding[dv] =  {'zlib':True, \
                            'complevel':5,\
                            'shuffle':True,\
                            '_FillValue':netcdf_fill_value}

    # ... coordinate encoding directives
    print('\n... creating coordinate encodings')
    coord_encoding = dict()
    for coord in ecco_ds.coords:
        # set default no fill value for coordinate
        if output_array_precision == np.float32:
            coord_encoding[coord] = {'_FillValue':None, 'dtype':'float32'}
        elif output_array_precision == np.float64:
            coord_encoding[coord] = {'_FillValue':None, 'dtype':'float64'}

        # force 64 bit ints to be 32 bit ints
        if (ecco_ds[coord].values.dtype == np.int32) or \
           (ecco_ds[coord].values.dtype == np.int64) :
            coord_encoding[coord]['dtype'] ='int32'

        # fix encoding of time
        if coord == 'time' or coord == 'time_bnds':
            coord_encoding[coord]['dtype'] ='int32'

            if 'units' in ecco_ds[coord].attrs:
                # apply units as encoding for time
                coord_encoding[coord]['units'] = ecco_ds[coord].attrs['units']
                # delete from the attributes list
                del ecco_ds[coord].attrs['units']

        elif coord == 'time_step':
            coord_encoding[coord]['dtype'] ='int32'

    # ... combined data variable and coordinate encoding directives
    encoding = {**dv_encoding, **coord_encoding}


    # Create directory for output files
    # ---------------------------------------------
    filepath = output_dir  / dataset_name

    if not filepath.exists():
        filepath.mkdir(parents=True, exist_ok=True)


    # Determine output freqency code.
    # ---------------------------------------------
    # user can specify directory or it can be found if the dataset
    # has the 'time_coverage_resolution' global attribute
    if output_freq_code == None:
        if 'time_coverage_resolution' in ecco_ds.attrs:

            print('dataset time averaging from metadata')
            time_coverage_resolution = ecco_ds.attrs['time_coverage_resolution']
            if time_coverage_resolution == 'P1M':
                output_freq_code='AVG_MON'
            elif time_coverage_resolution == 'P1D':
                output_freq_code='AVG_DAY'
            elif time_coverage_resolution == 'P0S':
                output_freq_code='SNAP'
        else:
            print('output_freq_code not defined and not available in dataset metadata')
            print('... using full record time in filename')


    # Write records to disk as NetCDF
    # ---------------------------------------------
    # one file per time level

    if time_method == 'by_record':
        for time_i, rec_time in enumerate(ecco_ds.time):

            cur_ds = ecco_ds.isel(time=time_i)

            # cast data variables to desired precision (if necessary)
            #for data_var in cur_ds.data_vars:
            #    if cur_ds[data_var].values.dtype != output_array_precision:
            #        cur_ds[data_var].values = cur_ds[data_var].astype(output_array_precision)

            time_date_info  =\
                make_date_str_from_dt64(cur_ds.time.values, output_freq_code)

           # sort comments alphabetically
            print('\n... sorting global attributes')
            cur_ds.attrs = sort_attrs(cur_ds.attrs)

            # add one final comment (PODAAC request)
            cur_ds.attrs["coordinates_comment"] = \
                "Note: the global 'coordinates' attribute descibes auxillary coordinates."

            fname = dataset_name + '_' + time_date_info['short'] +\
                    '_' + time_date_info['ppp_tttt'] + '.nc'

            print(fname)
            print(cur_ds)
            netcdf_output_filename = filepath / fname

            # SAVE
            print('\n... saving to netcdf ', netcdf_output_filename)
            cur_ds.to_netcdf(netcdf_output_filename, encoding=encoding)
            cur_ds.close()

    # one file per year
    elif time_method == 'by_year':
        unique_years = np.unique(ecco_ds.time.dt.year)
        print(unique_years)

        for year in unique_years:
            # pull out only records for this year
            cur_ds = ecco_ds.sel(time=slice(str(year), str(year)))

            first_time = cur_ds.time.values[0]
            last_time = cur_ds.time.values[-1]

            first_time_date_info =\
                make_date_str_from_dt64(first_time, output_freq_code)

            last_time_date_info =\
                make_date_str_from_dt64(last_time, output_freq_code)

           # sort comments alphabetically
            print('\n... sorting global attributes')
            cur_ds.attrs = sort_attrs(cur_ds.attrs)

            # add one final comment (PODAAC request)
            cur_ds.attrs["coordinates_comment"] = \
                "Note: the global 'coordinates' attribute descibes auxillary coordinates."

            fname = dataset_name + '_' +\
                first_time_date_info['short'] + '_' +\
                last_time_date_info['short'] + '_' +\
                first_time_date_info['ppp_tttt']+ '.nc'

            print(fname)
            print(cur_ds)
            netcdf_output_filename = filepath / fname

            # SAVE
            print('\n... saving to netcdf ', netcdf_output_filename)
            cur_ds.to_netcdf(netcdf_output_filename, encoding=encoding)
            cur_ds.close()


#%%%%
def make_date_str_from_dt64(dt64_time, output_freq_code):
    """Extracts components of a numpy date time 64 object

    Parameters
    ----------
    dt_64_time: numpy.datetime64
        a single datetime64 object

    output_freq_code: string
        a string code indicating the time level of averaging of the
        data variables
        options include
            'AVG_MON'  - monthly-averaged file
            'AVG_DAY'  - daily-averaged files
            'SNAPSHOT' - snapshot files (instantaneous)

    RETURNS:
    ----------
    a dictionary with the following string entries (all zero padded)
        date_str_full  : YYYY-MM-DDTHHMMSS
        date_str_short : YYYY-MM    (for AVG_MON)
                         YYYY-MM-DD (for AVG_DAY)
                         YYYY-MM-DDTHHMMSS (for SNAP)
        year           : YYYY
        month          : MM
        day            : DD
        hour           : HH
        ppp_tttt  : one of 'mon_mean','day_mean','snap'

    """

    print(dt64_time)
    date_str_full = str(dt64_time)[0:19].replace(':','')
    year  = date_str_full[0:4]
    month = date_str_full[5:7]
    day   = date_str_full[8:10]
    hour  = date_str_full[11:13]

    print(year, month, day, hour)
    ppp_tttt = ""
    date_str_short =""

    if output_freq_code == 'AVG_MON':
        date_str_short = str(np.datetime64(dt64_time,'M'))
        ppp_tttt = 'mon_mean'

    # --- AVG DAY
    elif output_freq_code == 'AVG_DAY':
        date_str_short = str(np.datetime64(dt64_time,'D'))
        ppp_tttt = 'day_mean'

    # --- SNAPSHOT
    elif 'SNAP' in output_freq_code:
        # convert from oroginal
        #   '1992-01-16T12:00:00.000000000'
        # to new format
        # '1992-01-16T120000'
        date_str_short = str(dt64_time)[0:19].replace(':','')
        ppp_tttt = 'snap'

    date_str = dict()
    date_str['full'] = date_str_full
    date_str['short'] = date_str_short
    date_str['year']  = year
    date_str['month'] = month
    date_str['day'] = day
    date_str['hour'] = hour
    date_str['ppp_tttt'] = ppp_tttt

    return date_str


#%%
def create_nc_grid_files_on_native_grid_from_mds(grid_input_dir,
                                                 grid_output_dir,
                                                 coordinate_metadata = dict(),
                                                 geometry_metadata = dict(),
                                                 global_metadata = dict(),
                                                 cell_bounds = None,
                                                 file_basename='ECCO-GRID',
                                                 title='llc grid geometry',
                                                 mds_datatype = '>f4',
                                                 write_to_disk = True,
                                                 less_output=True):

    # grid input dir and grid output dir should be of type pathlib.PosixPath

    mds_files = ''

    if isinstance(grid_input_dir, str):
        grid_input_dir = Path(grid_input_dir)
    if isinstance(grid_output_dir, str):
        grid_output_dir = Path(grid_output_dir)

    if not less_output:
        print(str(grid_input_dir))

    grid =  load_ecco_vars_from_mds(str(grid_input_dir),
                                    str(grid_input_dir),
                                    mds_files,
                                    vars_to_load = 'all',
                                    drop_unused_coords = False,
                                    grid_vars_to_coords = False,
                                    coordinate_metadata = coordinate_metadata,
                                    variable_metadata = geometry_metadata,
                                    global_metadata = global_metadata,
                                    cell_bounds = cell_bounds,
                                    mds_datatype = mds_datatype,
                                    less_output = less_output)

    if not less_output:
        print(grid)

#    for key in grid.attrs.keys():
#        if 'geo' in key or 'time' in key or 'cell_method' in key:
#            grid.attrs.pop(key)


    if 'time_coverage_end' in list(grid.attrs.keys()):
        grid.attrs.pop('time_coverage_end')

    if 'time_coverage_start' in list(grid.attrs.keys()):
        grid.attrs.pop('time_coverage_start')


    grid.attrs['title'] = title
    grid.attrs['product_name'] = file_basename + '.nc'
    grid.attrs['uuid'] = str(uuid.uuid1())


    # remove attributes that have TBD in their values
    for attr in list(grid.attrs.keys()):
        if type(grid.attrs[attr]) == str :
            if 'TBD' in grid.attrs[attr]:
                grid.attrs.pop(attr)

    # save to disk
    if write_to_disk:
        if not grid_output_dir.exists():
            try:
                grid_output_dir.mkdir()
            except:
                print ('cannot make %s ' % grid_output_dir)

        new_fname = grid_output_dir / (file_basename + '.nc')

        if not less_output:
            print('making single file grid netcdf')
            print(str(new_fname))

        if not less_output:
            print('\n... creating variable encodings')

        # PROVIDE SPECIFIC ENCODING DIRECTIVES FOR EACH DATA VAR
        dv_encoding = dict()
        for dv in grid.data_vars:
            dv_encoding[dv] =  {'zlib':True, \
                                'complevel':5,\
                                'shuffle':True,\
                                '_FillValue':None}


        # PROVIDE SPECIFIC ENCODING DIRECTIVES FOR EACH COORDINATE
        if not less_output:
            print('\n... creating coordinate encodings')
        coord_encoding = dict()

        for coord in grid.coords:
            # default encoding: no fill value, float32
            coord_encoding[coord] = {'_FillValue':None, 'dtype':'float32'}

            if grid[coord].values.dtype == np.int64:
                grid[coord].values[:] = grid[coord].astype(np.int32)
                coord_encoding[coord]['dtype'] ='int32'

        # MERGE ENCODINGS for coordinates and variables
        encoding = {**dv_encoding, **coord_encoding}


        if not less_output:
            print('\n... saving single file to netcdf')
        grid.to_netcdf(str(new_fname), encoding=encoding)

        # save as 13 tiles
        if not less_output:
            print('making 13 file grid netcdf')

        for i in range(13):
            tmp = grid.sel(tile=i)
            tmp2 = file_basename  + '_TILE_' + str(i).zfill(2) + '.nc'
            tmp.attrs['product_name'] = tmp2

            new_fname = grid_output_dir  / tmp2
            if not less_output:
                print (new_fname)

            tmp.to_netcdf(str(new_fname), encoding=encoding)

    return grid



#%%
def get_time_steps_from_mds_files(mds_var_dir, mds_file, less_output=True):

    if isinstance(mds_var_dir, str):
        mds_var_dir = Path(mds_var_dir)

    tmp_files = np.sort(list(mds_var_dir.glob(mds_file + '*meta')))

    time_steps = []

    print ('get time steps')
    if not less_output:
        print(tmp_files)
    for i in range(len(tmp_files)):
        time_steps.append(int(tmp_files[i].stem[-10:]))

    return time_steps
#%%


# create the interpolated fields. Default is on 0.5 degrees by 0.5 degrees.
def create_nc_variable_files_on_regular_grid_from_mds(mds_var_dir,
                                                     mds_files_to_load,
                                                     mds_grid_dir,
                                                     output_dir,
                                                     output_freq_code,
                                                     vars_to_load = 'all',
                                                     tiles_to_load = [0,1,2,3,4,5,6,7,8,9,10,11,12],
                                                     time_steps_to_load = [],
                                                     meta_variable_specific = dict(),
                                                     meta_common = dict(),
                                                     mds_datatype = '>f4',
                                                     dlon=0.5, dlat=0.5,
                                                     radius_of_influence = 120000,
                                                     express=1,
                                                     kvarnmidx = 2, # coordinate idx for vertical axis
                                                     # method now is only a place holder.
                                                     # This can be expanded. For example,
                                                     # the global interpolated fields can
                                                     # split to tiles, similarly to
                                                     # the tiled native fields, to
                                                     # reduce the size of each file.
                                                     verbose=True,
                                                     method = ''):
    #%%
    # force mds_files_to_load to be a list (if str is passed)
    if isinstance(mds_files_to_load, str):
        mds_files_to_load = [mds_files_to_load]

    # force time_steps_to_load to be a list (if int is passed)
    if isinstance(time_steps_to_load, int):
        time_steps_to_load = [time_steps_to_load]

    # for ce tiles_to_load to be a list (if int is passed)
    if isinstance(tiles_to_load, int):
        tiles_to_load = [tiles_to_load]

    # if no specific file data passed, read default metadata from json file
    # -- variable specific meta data
    script_dir = os.path.dirname(__file__)  # <-- absolute dir the script is in
    if not meta_variable_specific:
        meta_variable_rel_path = '../meta_json/ecco_meta_variable.json'
        abs_meta_variable_path = os.path.join(script_dir, meta_variable_rel_path)
        with open(abs_meta_variable_path, 'r') as fp:
            meta_variable_specific = json.load(fp)

    # --- common meta data
    if not meta_common:
        meta_common_rel_path = '../meta_json/ecco_meta_common.json'
        abs_meta_common_path = os.path.join(script_dir, meta_common_rel_path)
        with open(abs_meta_common_path, 'r') as fp:
            meta_common = json.load(fp)

    # info for the regular grid
    new_grid_min_lat = -90+dlat/2.
    new_grid_max_lat = 90-dlat/2.
    new_grid_min_lon = -180+dlon/2.
    new_grid_max_lon = 180-dlon/2.
    new_grid_ny = np.int((new_grid_max_lat-new_grid_min_lat)/dlat + 1 + 1e-4*dlat)
    new_grid_nx = np.int((new_grid_max_lon-new_grid_min_lon)/dlon + 1 + 1e-4*dlon)
    j_reg = new_grid_min_lat + np.asarray(range(new_grid_ny))*dlat
    i_reg = new_grid_min_lon + np.asarray(range(new_grid_nx))*dlon
    j_reg_idx = np.asarray(range(new_grid_ny))
    i_reg_idx = np.asarray(range(new_grid_nx))
    if (new_grid_ny < 1) or (new_grid_nx < 1):
        raise ValueError('You need to have at least one grid point for the new grid.')

    # loop through each mds file in mds_files_to_load
    for mds_file in mds_files_to_load:

        # if time steps to load is empty, load all time steps
        if len(time_steps_to_load ) == 0:
            # go through each file, pull out the time step, add the time step to a list,
            # and determine the start and end time of each record.

           time_steps_to_load = \
               get_time_steps_from_mds_files(mds_var_dir, mds_file)


        first_meta_fname  = mds_file + '.' + \
            str(time_steps_to_load[0]).zfill(10) + '.meta'


        # get metadata for the first file and determine which variables
        # are present
        meta = xm.utils.parse_meta_file(mds_var_dir + '/' + first_meta_fname)
        vars_here =  meta['fldList']

        if not isinstance(vars_to_load, list):
            vars_to_load = [vars_to_load]

        if 'all' not in vars_to_load:
            num_vars_matching = len(np.intersect1d(vars_to_load, vars_here))

            print ('num vars matching ', num_vars_matching)

            # only proceed if we are sure that the variable we want is in this
            # mds file
            if num_vars_matching == 0:
                print ('none of the variables you want are in ', mds_file)
                print (vars_to_load)
                print (vars_here)

                break
        #%%
        # load the MDS fields
        ecco_dataset_all =  \
                load_ecco_vars_from_mds(mds_var_dir, \
                                         mds_grid_dir,
                                         mds_file,
                                         vars_to_load = vars_to_load,
                                         tiles_to_load=tiles_to_load,
                                         model_time_steps_to_load=time_steps_to_load,
                                         output_freq_code = \
                                              output_freq_code,
                                         meta_variable_specific = \
                                              meta_variable_specific,
                                         meta_common=meta_common,
                                         mds_datatype=mds_datatype,
                                         llc_method = 'bigchunks')

        # do the actual loading. Otherwise, the code may be slow.
        ecco_dataset_all.load()

        # print(ecco_dataset_all.keys())
        # loop through each variable in this dataset,
        for var in ecco_dataset_all.keys():
            print ('    ' + var)
            # obtain the grid information (use fields from time=0)
            # Note that nrtmp would always equal to one,
            # since each outfile will include only one time-record (e.g. daily, monthly avgs.).

            ecco_dataset = ecco_dataset_all.isel(time=[0])

            var_ds = ecco_dataset[var]

            shapetmp = var_ds.shape

            lenshapetmp = len(shapetmp)
            nttmp = 0
            nrtmp = 0
            if(lenshapetmp==4):
                nttmp = shapetmp[0]
                nrtmp = 0
            elif(lenshapetmp==5):
                nttmp = shapetmp[0]
                nrtmp = shapetmp[1]
            else:
                print('Error! ', var_ds.shape)
                sys.exit()

            # Get X,Y of the original grid. They could be XC/YC, XG/YC, XC/YG, etc.
            # Similar for mask.
            # default is XC, YC
            if 'i' in var_ds.coords.keys():
                XX = ecco_dataset['XC']
                XXname = 'XC'
            if 'j' in var_ds.coords.keys():
                YY = ecco_dataset['YC']
                YYname = 'YC'
            varmask = 'maskC'
            iname = 'i'
            jname = 'j'

            if 'i_g' in var_ds.coords.keys():
                XX = ecco_dataset['XG']
                XXname = 'XG'
                varmask = 'maskW'
                iname = 'i_g'
            if 'j_g' in var_ds.coords.keys():
                YY = ecco_dataset['YG']
                YYname = 'YG'
                varmask = 'maskS'
                jname = 'j_g'

            # interpolation
            # To do it fast, set express==1 (default)
            if(express==1):
                orig_lons_1d = XX.values.ravel()
                orig_lats_1d = YY.values.ravel()
                orig_grid = pr.geometry.SwathDefinition(lons=orig_lons_1d,
                                                        lats=orig_lats_1d)

                if (new_grid_ny > 0) and (new_grid_nx > 0):
                    # 1D grid values
                    new_grid_lon, new_grid_lat = np.meshgrid(i_reg, j_reg)

                    # define the lat lon points of the two parts.
                    new_grid  = pr.geometry.GridDefinition(lons=new_grid_lon,
                                                           lats=new_grid_lat)

                    # Get the neighbor info once.
                    # It will be used repeatedly late to resample data
                    # fast for each of the datasets that is based on
                    # the same swath, e.g. for a model variable at different times.
                    valid_input_index, valid_output_index, index_array, distance_array = \
                    pr.kd_tree.get_neighbour_info(orig_grid,
                                               new_grid, radius_of_influence,
                                               neighbours=1)

            # loop through time steps, one at a time.
            for time_step in time_steps_to_load:

                i, = np.where(ecco_dataset_all.timestep == time_step)
                if(verbose):
                    print (ecco_dataset_all.timestep.values)
                    print ('time step ', time_step, i)

                # load the dataset
                ecco_dataset = ecco_dataset_all.isel(time=i)

                # pull out the year, month day, hour, min, sec associated with
                # this time step
                if type(ecco_dataset.time.values) == np.ndarray:
                    cur_time = ecco_dataset.time.values[0]
                else:
                    cur_time = ecco_dataset.time.values
                #print (type(cur_time))
                year, mon, day, hh, mm, ss  = \
                     extract_yyyy_mm_dd_hh_mm_ss_from_datetime64(cur_time)

                print(year, mon, day)

                # if the field comes from an average,
                # extract the time bounds -- we'll use it before we save
                # the variable
                if 'AVG' in output_freq_code:
                    tb = ecco_dataset.time_bnds
                    tb.name = 'tb'

                var_ds = ecco_dataset[var]

                # 3d fields (with Z-axis) for each time record
                if(nttmp != 0 and nrtmp != 0):
                    tmpall = np.zeros((nttmp, nrtmp,new_grid_ny,new_grid_nx))
                    for ir in range(nrtmp): # Z-loop
                        # mask
                        maskloc = ecco_dataset[varmask].values[ir,:]

                        for it in range(nttmp): # time loop
                            # one 2d field at a time
                            var_ds_onechunk = var_ds[it,ir,:]
                            # apply mask
                            var_ds_onechunk.values[maskloc==0]=np.nan
                            orig_field = var_ds_onechunk.values
                            if(express==1):
                                tmp = pr.kd_tree.get_sample_from_neighbour_info(
                                        'nn', new_grid.shape, orig_field,
                                        valid_input_index, valid_output_index,
                                        index_array)

                            else:
                                new_grid_lon, new_grid_lat, tmp = resample_to_latlon(XX, YY, orig_field,
                                                                  new_grid_min_lat,
                                                                  new_grid_max_lat, dlat,
                                                                  new_grid_min_lon,
                                                                  new_grid_max_lon, dlon,
                                                                  nprocs_user=1,
                                                                  mapping_method = 'nearest_neighbor',
                                                                  radius_of_influence=radius_of_influence)
                            tmpall[it,ir,:] = tmp
                # 2d fields (without Z-axis) for each time record
                elif(nttmp != 0):
                    tmpall = np.zeros((nttmp, new_grid_ny,new_grid_nx))
                    # mask
                    maskloc = ecco_dataset[varmask].values[0,:]
                    for it in range(nttmp): # time loop
                        var_ds_onechunk = var_ds[it,:]
                        var_ds_onechunk.values[maskloc==0]=np.nan
                        orig_field = var_ds_onechunk.values
                        if(express==1):
                            tmp = pr.kd_tree.get_sample_from_neighbour_info(
                                    'nn', new_grid.shape, orig_field,
                                    valid_input_index, valid_output_index,
                                    index_array)
                        else:
                            new_grid_lon, new_grid_lat, tmp = resample_to_latlon(XX, YY, orig_field,
                                                              new_grid_min_lat,
                                                              new_grid_max_lat, dlat,
                                                              new_grid_min_lon,
                                                              new_grid_max_lon, dlon,
                                                              nprocs_user=1,
                                                              mapping_method = 'nearest_neighbor',
                                                              radius_of_influence=radius_of_influence)
                        tmpall[it,:] = tmp

                else:
                    print('Error! both nttmp and nrtmp are zeros.')
                    sys.exit()
                # set the coordinates for the new (regular) grid
                # 2d fields
                if(lenshapetmp==4):
                    var_ds_reg = xr.DataArray(tmpall,
                                              coords = {'time': var_ds.coords['time'].values,
                                                        'j': j_reg_idx,
                                                        'i': i_reg_idx},\
                                              dims = ('time', 'j', 'i'))
                # 3d fields
                elif(lenshapetmp==5):
                    # Get the variable name (kvarnm) for Z-axis: k, k_l
                    kvarnm = var_ds.coords.keys()[kvarnmidx]

                    if(kvarnm[0]!='k'):
                        kvarnmidxnew = kvarnmidx
                        for iktmp, ktmp in enumerate(var_ds.coords.keys()):
                            if(ktmp[0]=='k'):
                                kvarnmidxnew = iktmp
                        if(kvarnmidxnew==kvarnmidx):
                            print('Error! Seems ', kvarnm, ' is not the vertical axis.')
                            print(var_ds)
                            sys.exit()
                        else:
                            kvarnmidx = kvarnmidxnew
                            kvarnm = var_ds.coords.keys()[kvarnmidx]

                    var_ds_reg = xr.DataArray(tmpall,
                                              coords = {'time': var_ds.coords['time'].values,
                                                        kvarnm: var_ds.coords[kvarnm].values,
                                                        'j': j_reg_idx,
                                                        'i': i_reg_idx},\
                                              dims = ('time', kvarnm,'j', 'i'))
                # set the attrs for the new (regular) grid
                var_ds_reg['j'].attrs = var_ds[jname].attrs
                var_ds_reg['i'].attrs = var_ds[iname].attrs
                var_ds_reg['j'].attrs['long_name'] = 'y-dimension'
                var_ds_reg['i'].attrs['long_name'] = 'x-dimension'
                var_ds_reg['j'].attrs['swap_dim'] = 'latitude'
                var_ds_reg['i'].attrs['swap_dim'] = 'longitude'

                var_ds_reg['latitude'] = (('j'), j_reg)
                var_ds_reg['longitude'] = (('i'), i_reg)
                var_ds_reg['latitude'].attrs = ecco_dataset[YYname].attrs
                var_ds_reg['longitude'].attrs = ecco_dataset[XXname].attrs
                var_ds_reg['latitude'].attrs['long_name'] = "latitude at center of grid cell"
                var_ds_reg['longitude'].attrs['long_name'] = "longitude at center of grid cell"

                var_ds_reg.name = var_ds.name

                #keys_to_drop = ['tile','j','i','XC','YC','XG','YG']
                # drop these ancillary fields -- they are in grid anyway
                keys_to_drop = ['CS','SN','Depth','rA','PHrefC','hFacC',\
                                'maskC','drF', 'dxC', 'dyG', 'rAw', 'hFacW',\
                                'rAs','hFacS','maskS','dxG','dyC', 'maskW', \
                                'tile','XC','YC','XG','YG']

                for key_to_drop in keys_to_drop:
                    #print (key_to_drop)
                    if key_to_drop in var_ds.coords.keys():
                        var_ds = var_ds.drop(key_to_drop)

                # any remaining fields, e.g. time, would be included in the interpolated fields.
                for key_to_add in var_ds.coords.keys():
                    if(key_to_add not in var_ds_reg.coords.keys()):
                        if(key_to_add != 'i_g' and key_to_add != 'j_g'):
                            var_ds_reg[key_to_add] = var_ds[key_to_add]

                # use the same global attributs
                var_ds_reg.attrs = var_ds.attrs


                #print(var_ds.coords.keys())
                #%%

                # create the new file path name
                if 'MON' in output_freq_code:

                    fname = var + '_' +  str(year) + '_' + str(mon).zfill(2) + '.nc'

                    newpath = output_dir + '/' + var + '/' + \
                        str(year) + '/'

                elif ('WEEK' in output_freq_code) or \
                     ('DAY' in output_freq_code):

                    fname = var + '_' + \
                            str(year) + '_' + \
                            str(mon).zfill(2) + '_' + \
                            str(day).zfill(2) +  '.nc'
                    d0 = datetime.datetime(year, 1,1)
                    d1 = datetime.datetime(year, mon, day)
                    doy = (d1-d0).days + 1

                    newpath = output_dir + '/' + var + '/' + \
                        str(year) + '/' + str(doy).zfill(3)

                elif 'YEAR' in output_freq_code:

                     fname = var + '_' + str(year) + '.nc'

                     newpath = output_dir + '/' + var + '/' + \
                        str(year)

                else:
                    print ('no valid output frequency code specified')
                    print ('saving to year/mon/day/tile')
                    fname = var + '_' + \
                        str(year) + '_' + \
                        str(mon).zfill(2) + '_' + \
                        str(day).zfill(2) + '.nc'
                    d0 = datetime.datetime(year, 1,1)
                    d1 = datetime.datetime(year, mon, day)
                    doy = (d1-d0).days + 1

                    newpath = output_dir + '/' + var + '/' + \
                        str(year) + '/' + str(doy).zfill(3)


                # create the path if it does not exist/
                if not os.path.exists(newpath):
                    os.makedirs(newpath)

                # convert the data array to a dataset.
                tmp = var_ds_reg.to_dataset()

                # add the time bounds field back in if we have an
                # average field
                if 'AVG' in output_freq_code:
                    tmp = xr.merge((tmp, tb))
                    tmp = tmp.drop('tb')

                # put the metadata back in
                tmp.attrs = ecco_dataset.attrs

                # update the temporal and geospatial metadata
                tmp = update_ecco_dataset_geospatial_metadata(tmp)
                tmp = update_ecco_dataset_temporal_coverage_metadata(tmp)

                # save to netcdf.  it's that simple.
                if(verbose):
                    print ('saving to %s' % newpath + '/' + fname)
                # do not include _FillValue
                encoding = {i: {'_FillValue': False} for i in tmp.variables.keys()}

                tmp.to_netcdf(newpath + '/' + fname, engine='netcdf4',encoding=encoding)



#%%
    ecco_dataset_all.close()
    return ecco_dataset, tmp






#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
ECCO v4 Python: Utililites

This module includes utility routines that operate on the Dataset or DataArray Objects

.. _ecco_v4_py Documentation :
   https://github.com/ECCO-GROUP/ECCOv4-py
"""

from __future__ import division, print_function
import numpy as np
import pandas as pd
import xarray as xr
import datetime
import dateutil
import xgcm
from pprint import pprint
from collections import OrderedDict

def find_metadata_in_json_dictionary(var, key, metadata, print_output=False):
    for m in metadata:
        if key in m.keys():
            if m[key] == var:
                if print_output:
                    print(m)
                return m

    return []



def add_global_metadata(metadata, G, dataset_dim, less_output=True):

    if not less_output:
        print('adding global metadata')
        pprint(metadata)

    # loop through pairs
    for mc in metadata:
        # get name and type
        mname = mc['name']
        mtype = mc['type']

        # by default add the key/pair
        add_field = True

        # unless it has as specific 'grid_dimension' associated
        # with it. If so, then only add it if this dataset indicates
        # it is necessary.  for example, don't provide geospatial
        # depth information for 2D datasets
        if 'grid_dimension' in mc.keys():
            gd = mc['grid_dimension']

            if dataset_dim not in gd:
                add_field = False

        # if we do add the field, we have to convert to the
        # appropriate data type
        if add_field == True:
            if isinstance(mc['value'], str) and 'TBD_' in mc['value']:
                G.attrs[mname] = mc['value']
            else:
                if mtype == 's':
                    G.attrs[mname] = mc['value']
                elif mtype == 'f':
                    G.attrs[mname] = float(mc['value'])
                elif mtype == 'i':
                    G.attrs[mname] = np.int32(mc['value'])
                else:
                    print('INVALID MTYPE ! ', mtype)
        else:
            if not less_output:
                print('\t> not adding ', mc)

    return G


def add_coordinate_metadata(metadata_dict, G, less_output=True):
    # G : dataset
    # metadata_dict: dictionary of metadata records with name as a key
    keys_to_exclude = ['grid_dimension','name']

    for coord in G.coords:

        if not less_output:
            print('\n### ', coord)
        # look for coordinate in metadat dictionary
        mv = find_metadata_in_json_dictionary(coord, 'name', metadata_dict)

        if len(mv) > 0:
            # if metadata for this coordinate is present
            # loop through all of the keys and if it is not
            # on the excluded list, add it
            for m_key in sorted(mv.keys()):
                if m_key not in keys_to_exclude:
                    G[coord].attrs[m_key] = mv[m_key]

                    if not less_output:
                        print('\t',m_key, ':', mv[m_key])
        else:
            if not less_output:
                print('...... no metadata found in dictionary')

    return G


def add_variable_metadata(variable_metadata_dict, G, \
                          grouping_gcmd_keywords=[], less_output=True):

    # ADD VARIABLE METADATA  & SAVE GCMD KEYWORDS
    keys_to_exclude = ['grid_dimension','name', 'GCMD_keywords', "variable_rename",\
                       'comments_1', 'comments_2', 'internal_note','internal note',\
                       'grid_location']

    for var in G.data_vars:
        if not less_output:
            print('\n### ', var)

        mv = find_metadata_in_json_dictionary(var, 'name', variable_metadata_dict)

        if len(mv) == 0:
            if not less_output:
                print('...... no metadata found in dictionary')

        else:
            # loop through each key, add if not on exclude list
            for m_key in sorted(mv.keys()):
                if m_key not in keys_to_exclude:
                    G[var].attrs[m_key] = mv[m_key]
                    if not less_output:
                        print('\t',m_key, ':', mv[m_key])

            # merge the two comment fields (both *MUST* be present in the json file
            # but they can be empty strings "")

            # if either one is not-empty, add a comment field
            if len(mv['comments_1']) > 0 or len(mv['comments_2']) > 0:

                # if both are not-empty, merge them, make sure '.' between
                if len(mv['comments_1']) > 0 and len(mv['comments_2']) > 0:
                    if mv['comments_1'][-1] == '.':
                        G[var].attrs['comment'] =  mv['comments_1'] + ' ' + mv['comments_2']
                    else:
                        G[var].attrs['comment'] =  mv['comments_1'] + '. ' + mv['comments_2']

                # if only comments_1 is not empty
                elif len(mv['comments_1']) > 0:
                    G[var].attrs['comment'] = mv['comments_1']

                # if only comments_2 is not empty
                elif len(mv['comments_2']) > 0:
                   G[var].attrs['comment'] = mv['comments_2']

                if not less_output:
                    print('\t', 'comment', ':', G[var].attrs['comment'])
            else:
                if not less_output:
                    print('\t', 'comment fields are empty')

            # append GCMD keywords, if present
            if 'GCMD_keywords' in mv.keys():
               # Get the GCMD keywords, these will be added into the global
               # attributes
               gcmd_keywords = mv['GCMD_keywords'].split(',')

               if not less_output:
                   print('\t','GCMD keywords : ', gcmd_keywords)

               for gcmd_keyword in gcmd_keywords:
                   grouping_gcmd_keywords.append(gcmd_keyword.strip())

    return G, grouping_gcmd_keywords


def sort_attrs(attrs):
    """

    Alphabetically sort all keys in a dictionary

    Parameters
    ----------
    attrs : dict
        a dictionary of key/value pairs

    Returns
    -------
    attrs : dict
        a dictionary of key/value pairs sorted alphabetically by key

    """

    od = OrderedDict()

    keys = sorted(list(attrs.keys()),key=str.casefold)

    for k in keys:
        od[k] = attrs[k]

    return od


def sort_all_attrs(ecco_dataset, print_output=False):
    """

    Alphabetically sort all attributes in a ecco_dataset object (ecco_dataset)
    including the attributes of coordinates, data variables,
    and the global attributes.

    Parameters
    ----------
    ecco_dataset : xarray Dataset
        an xarray dataset

    Returns
    -------
    ecco_dataset : xarray Dataset
        an xarray dataset with all attributes sorted alphabetically

    """

    for coord in list(ecco_dataset.coords):
        if print_output:
            print(coord)
        new_attrs = sort_attrs(ecco_dataset[coord].attrs)
        ecco_dataset[coord].attrs = new_attrs

    for dv in list(ecco_dataset.data_vars):
        if print_output:
            print(dv)
        new_attrs = sort_attrs(ecco_dataset[dv].attrs)
        ecco_dataset[dv].attrs = new_attrs

    new_attrs = sort_attrs(ecco_dataset.attrs)
    ecco_dataset.attrs = new_attrs

    return ecco_dataset


def make_time_bounds_and_center_times_from_ecco_dataset(ecco_dataset, \
                                                        output_freq_code):
    """

    Given an ecco_dataset object (ecco_dataset) with time variables that
    correspond with the 'end' of averaging time periods
    and an output frequency code (AVG_MON, AVG_DAY, AVG_WEEK, or AVG_YEAR),
    create a time_bounds array of dimension 2xn, with 'n' being the number
    of time averaged records in ecco_dataset, each with two datetime64
    variables, one for the averaging period start, one
    for the averaging period end.

    The routine also creates an array of times corresponding to the 'middle'
    of each averaging period.

    Parameters
    ----------
    ecco_dataset : xarray Dataset
        an xarray dataset with 'time' variables representing the times at
        the 'end' of an averaging periods


    Returns
    -------
    time_bnds : np.array(dtype=np.datetime64)
        a datetime64 with the start and end time(s) of the averaging periods

    center_times :np.array(dtype=np.datetime64)
        a numpy array containing the 'center' time of the averaging period(s)


    """

    if ecco_dataset.time.size == 1:

        if isinstance(ecco_dataset.time.values, np.ndarray):
            time_tmp = ecco_dataset.time.values[0]
        else:
            time_tmp = ecco_dataset.time.values

        time_bnds, center_times = \
            make_time_bounds_from_ds64(time_tmp,\
                                       output_freq_code)

        time_bnds=np.expand_dims(time_bnds,0)

    else:
        time_start = []
        time_end  = []
        center_time = []
        for time_i in range(len(ecco_dataset.timestep)):
             tb, ct = \
                 make_time_bounds_from_ds64(ecco_dataset.time.values[time_i],
                                  output_freq_code)

             time_start.append(tb[0])
             time_end.append(tb[1])

             center_time.append(ct)

        # convert list to array
        center_times = np.array(center_time,dtype=np.datetime64)
        time_bnds    = np.array([time_start, time_end],dtype='datetime64')
        time_bnds    = time_bnds.T

    # make time bounds dataset
    if 'time' not in ecco_dataset.dims:
         ecco_dataset = ecco_dataset.expand_dims(dim='time')

    #print ('-- tb shape ', time_bnds.shape)
    #print ('-- tb type  ', type(time_bnds))

    time_bnds_ds = xr.Dataset({'time_bnds': (['time','nv'], time_bnds)},
                               coords={'time':ecco_dataset.time}) #,

    return time_bnds_ds, center_times

#%%
def make_time_bounds_from_ds64(rec_avg_end, output_freq_code):
    """

    Given a datetime64 object (rec_avg_end) representing the 'end' of an
    averaging time period (usually derived from the mitgcm file's timestep)
    and an output frequency code
    (AVG_MON, AVG_DAY, AVG_WEEK, or AVG_YEAR), create a time_bounds array
    with two datetime64 variables, one for the averaging period start, one
    for the averaging period end.  Also find the middle time between the
    two..

    Parameters
    ----------
    rec_avg_end : numpy.datetime64
        the time at the end of an averaging period

    output_freq_code : str
        code indicating the time period of the averaging period
        - AVG_DAY, AVG_MON, AVG_WEEK, or AVG_YEAR


    Returns
    -------
    time_bnds : numpy.array(dtype=numpy.datetime64)
        a datetime64 array with the start and end time of the averaging periods

    center_times : numpy.datetime64
        the 'center' of the averaging period

    """

    if  output_freq_code in ('AVG_MON','AVG_DAY','AVG_WEEK','AVG_YEAR'):
        rec_year, rec_mon, rec_day, \
        rec_hour, rec_min, rec_sec = \
            extract_yyyy_mm_dd_hh_mm_ss_from_datetime64(rec_avg_end)


        rec_avg_end_as_dt = datetime.datetime(rec_year, rec_mon,
                                              rec_day, rec_hour,
                                              rec_min, rec_sec)

        if output_freq_code     == 'AVG_MON':
            rec_avg_start =  rec_avg_end_as_dt - \
                dateutil.relativedelta.relativedelta(months=1)
        elif output_freq_code   == 'AVG_DAY':
            rec_avg_start =  rec_avg_end_as_dt - \
                dateutil.relativedelta.relativedelta(days=1)
        elif output_freq_code   == 'AVG_WEEK':
            rec_avg_start =  rec_avg_end_as_dt - \
                dateutil.relativedelta.relativedelta(weeks=1)
        elif output_freq_code   == 'AVG_YEAR':
            rec_avg_start =  rec_avg_end_as_dt - \
                dateutil.relativedelta.relativedelta(years=1)

        rec_avg_start =  np.datetime64(rec_avg_start)

        rec_avg_delta = rec_avg_end - rec_avg_start
        rec_avg_middle = rec_avg_start + rec_avg_delta/2
        #print rec_avg_end, rec_avg_start, rec_avg_middle

        rec_time_bnds = np.array([rec_avg_start, rec_avg_end])

        return rec_time_bnds, rec_avg_middle

    else:
        print ('output_freq_code must be: AVG_MON, AVG_DAY, AVG_WEEK, OR AVG_YEAR')
        print ('you provided ' + str(output_freq_code))
        return [],[]

#%%
def extract_yyyy_mm_dd_hh_mm_ss_from_datetime64(dt64):
    """

    Extract separate fields for year, month, day, hour, min, sec from
    a datetime64 object, or an array-like object of datetime64 objects

    Parameters
    ----------
    dt64 : xarray DataArray, np.ndarray, list of, or single numpy.datetime64
        datetime64 object

    Returns
    -------
    year, mon, day, hh, mm, ss : int

    """
    # use xarray to do this robustly
    if isinstance(dt64,xr.core.dataarray.DataArray):
        year = dt64.dt.year.astype(int)
        mon = dt64.dt.month.astype(int)
        day = dt64.dt.day.astype(int)
        hh = dt64.dt.hour.astype(int)
        mm = dt64.dt.minute.astype(int)
        ss = dt64.dt.second.astype(int)
        return year, mon, day, hh, mm, ss


    # otherwise transform the problem to use xarray
    elif isinstance(dt64,list):
        xdates = extract_yyyy_mm_dd_hh_mm_ss_from_datetime64(xr.DataArray(np.array(dt64)))
        return tuple([list(x.values) for x in xdates])
    elif isinstance(dt64,np.ndarray):
        xdates = extract_yyyy_mm_dd_hh_mm_ss_from_datetime64(xr.DataArray(dt64))
        return tuple([x.values for x in xdates])
    elif isinstance(dt64,np.datetime64):
        xdates = extract_yyyy_mm_dd_hh_mm_ss_from_datetime64(xr.DataArray(pd.to_datetime(dt64)))
        return tuple([int(x.values) for x in xdates])

#%%
def minimal_metadata(ds):
    """

    This routine removes some of the redundant metadata that is included with the ECCO v4 netcdf tiles from the Dataset object `ds`.  Specifically, metadata with the tags `A` through `Z` (those metadata records) that describe the origin of the ECCO v4 output.

    Parameters
    ----------
    ds : xarray Dataset
        An `xarray` Dataset object that was created by loading an
        ECCO v4 tile netcdf file


    """

    print('Removing Dataset Attributes A-Z\n')
    # generate a list of upper case letters in teh alphabet
    myDict= map(chr, range(65, 91))

    for key, value in ds.attrs.items():
        if key in myDict:
            del ds.attrs[key]



#%%
def months2days(nmon=288, baseyear=1992, basemon=1):
    """

    This routine converts the mid-month time to days from January 1st of a particular year.

    Parameters
    ----------
    nmon : dtype=integer
        number of months

    baseyear : dtype=integer
        year of time of origin

    basemon : dtype=integer
        month of time of origin

    Returns
    -------
        time_days : ndarray
            the middle time of each month in days from Jan 1 baseyear (numpy array [nmon], dtype=double)

        time_days_bnds : ndarray
            time bounds (numpy array [nmon, 2], dtype=double)

        ansi_date : ndarray
            array of ANSI date strings

    """

    time_days_bnds = np.zeros([nmon,2])
    time_1stdayofmon = np.zeros([nmon+1])

    basetime = datetime.datetime(baseyear, basemon, 1, 0, 0, 0)

    for mon in range(nmon+1):
        #monfrombasemon is how many months fron basemon
        monfrombasemon=basemon+mon-1
        yrtmp = monfrombasemon//12+baseyear
        montmp = monfrombasemon%12+1
        tmpdate = datetime.datetime(yrtmp,montmp,1,0,0,0)-basetime
        time_1stdayofmon[mon] = tmpdate.days
    #time bounds are the 1st day of each month.
    time_days_bnds[:,0]= time_1stdayofmon[0:nmon]
    time_days_bnds[:,1]= time_1stdayofmon[1:nmon+1]
    #center time of each month is the mean of the time bounds.
    time_days = np.mean(time_days_bnds,axis=1)

    ansi_datetmp = np.array([basetime + datetime.timedelta(days=time_days[i]) for i in range(nmon)])
    ansi_date = [str.replace(ansi_datetmp[i].isoformat(),'T',' ') for i in range(nmon)]

    return time_days, time_days_bnds, ansi_date

#%%

def get_llc_grid(ds,domain='global'):
    """
    Define xgcm Grid object for the LLC grid
    See example usage in the xgcm documentation:
    https://xgcm.readthedocs.io/en/latest/example_eccov4.html#Spatially-Integrated-Heat-Content-Anomaly

    Parameters
    ----------
    ds : xarray Dataset
        formed from LLC90 grid, must have the basic coordinates:
        i,j,i_g,j_g,k,k_l,k_u,k_p1

    Returns
    -------
    grid : xgcm Grid object
        defines horizontal connections between LLC tiles

    """

    if 'domain' in ds.attrs:
        domain = ds.attrs['domain']

    if domain == 'global':
        # Establish grid topology
        tile_connections = {'tile':  {
                0: {'X': ((12, 'Y', False), (3, 'X', False)),
                    'Y': (None, (1, 'Y', False))},
                1: {'X': ((11, 'Y', False), (4, 'X', False)),
                    'Y': ((0, 'Y', False), (2, 'Y', False))},
                2: {'X': ((10, 'Y', False), (5, 'X', False)),
                    'Y': ((1, 'Y', False), (6, 'X', False))},
                3: {'X': ((0, 'X', False), (9, 'Y', False)),
                    'Y': (None, (4, 'Y', False))},
                4: {'X': ((1, 'X', False), (8, 'Y', False)),
                    'Y': ((3, 'Y', False), (5, 'Y', False))},
                5: {'X': ((2, 'X', False), (7, 'Y', False)),
                    'Y': ((4, 'Y', False), (6, 'Y', False))},
                6: {'X': ((2, 'Y', False), (7, 'X', False)),
                    'Y': ((5, 'Y', False), (10, 'X', False))},
                7: {'X': ((6, 'X', False), (8, 'X', False)),
                    'Y': ((5, 'X', False), (10, 'Y', False))},
                8: {'X': ((7, 'X', False), (9, 'X', False)),
                    'Y': ((4, 'X', False), (11, 'Y', False))},
                9: {'X': ((8, 'X', False), None),
                    'Y': ((3, 'X', False), (12, 'Y', False))},
                10: {'X': ((6, 'Y', False), (11, 'X', False)),
                     'Y': ((7, 'Y', False), (2, 'X', False))},
                11: {'X': ((10, 'X', False), (12, 'X', False)),
                     'Y': ((8, 'Y', False), (1, 'X', False))},
                12: {'X': ((11, 'X', False), None),
                     'Y': ((9, 'Y', False), (0, 'X', False))}
        }}

        grid = xgcm.Grid(ds,
                periodic=False,
                face_connections=tile_connections
        )
    elif domain == 'aste':
        tile_connections = {'tile':{
                    0:{'X':((5,'Y',False),None),
                       'Y':(None,(1,'Y',False))},
                    1:{'X':((4,'Y',False),None),
                       'Y':((0,'Y',False),(2,'X',False))},
                    2:{'X':((1,'Y',False),(3,'X',False)),
                       'Y':(None,(4,'X',False))},
                    3:{'X':((2,'X',False),None),
                       'Y':(None,None)},
                    4:{'X':((2,'Y',False),(5,'X',False)),
                       'Y':(None,(1,'X',False))},
                    5:{'X':((4,'X',False),None),
                       'Y':(None,(0,'X',False))}
                   }}
        grid = xgcm.Grid(ds,periodic=False,face_connections=tile_connections)
    else:
        raise TypeError(f'Domain {domain} not recognized')


    return grid



### This module contains routines to access and retrieve ECCO datasets on the AWS Cloud.
### These functions will only work when called from an AWS EC2 instance running in region us-west-2.

from .ecco_acc_dates import date_adjustment


## Initalize Python libraries for module
import numpy as np
import pandas as pd
import requests
import time as time
import os.path
from os.path import basename, isfile, isdir, join, expanduser
from pathlib import Path
from platform import system
from netrc import netrc
from urllib import request
from http.cookiejar import CookieJar
from getpass import getpass
import requests
import zarr
    

def setup_earthdata_login_auth(url: str='urs.earthdata.nasa.gov'):
    """Helper subroutine to log into NASA EarthData"""

    # Predict the path of the netrc file depending on os/platform type.
    _netrc = join(expanduser('~'), "_netrc" if system()=="Windows" else ".netrc")
    
    # look for the netrc file and use the login/password
    try:
        username, _, password = netrc(file=_netrc).authenticators(url)
    
    # if the file is not found, prompt the user for the login/password
    except (FileNotFoundError, TypeError):
        print('Please provide Earthdata Login credentials for access.')
        username, password = input('Username: '), getpass('Password: ')
        
        # write credentials to netrc file
        with open(_netrc,'a') as file:
            lines = ["machine urs.earthdata.nasa.gov\n",\
                     "    login "+username+"\n",\
                     "    password "+password]
            file.writelines(lines)
            file.close()
    
    manager = request.HTTPPasswordMgrWithDefaultRealm()
    manager.add_password(None, url, username, password)
    auth = request.HTTPBasicAuthHandler(manager)
    jar = CookieJar()
    processor = request.HTTPCookieProcessor(jar)
    opener = request.build_opener(auth, processor)
    request.install_opener(opener)



###================================================================================================================



def ecco_podaac_s3_query(ShortName,StartDate,EndDate,version,snapshot_interval='monthly'):
    
    """
    
    This routine searches for files of the given ShortName and date range.
    It returns a list of files that can be opened or downloaded to a user's local instance.
    This function is called by the other routines in this module.
    
    Parameters
    ----------
    ShortName: str, the ShortName that identifies the dataset on PO.DAAC.
    
    StartDate,EndDate: str, in 'YYYY', 'YYYY-MM', or 'YYYY-MM-DD' format, 
                       define date range [StartDate,EndDate] for download.
                       EndDate is included in the time range (unlike typical Python ranges).
                       ECCOv4r4 date range is '1992-01-01' to '2017-12-31'.
                       For 'SNAPSHOT' datasets, an additional day is added to EndDate to enable closed budgets
                       within the specified date range.

    version: ('v4r4','v4r5'), specifies ECCO version to query.
             Currently 'v4r5' only works with ['s3_open','s3_get','s3_get_ifspace'] modes,
             or if the files are already stored in download_root_dir/ShortName/.
             Otherwise an error is returned.
             'v4r5' only has grid='native' and time_res='monthly' data files available.
    
    snapshot_interval: ('monthly', 'daily'), if the dataset corresponding to ShortName is a snapshot, 
                       determines whether snapshots are included for only the beginning/end of each month 
                       ('monthly'), or for every day ('daily'). Defaults to 'monthly'.

    Returns
    -------
    s3_files_list: str or list, unopened file paths on S3 that match the query
    
    """

    pass


    ## Define Helper Subroutines
    
    ### Helper subroutines to make the API calls to search CMR and parse response
    def set_params(params: dict):
#         params.update({'scroll': "true", 'page_size': 2000})
        params.update({'page_size': 2000})
        return {par: val for par, val in params.items() if val is not None}
    
    def get_results(params: dict, headers: dict=None):
        response = requests.get(url="https://cmr.earthdata.nasa.gov/search/granules.json", 
                                params=set_params(params),
                                headers=headers).json()
        return response    
    
    def get_granules(params: dict, ShortName: str, SingleDay_flag: bool):
        time_start = np.array([]).astype('datetime64[ns]')
        s3_files_list = []
        completed_query = False
        while completed_query == False:
            response = get_results(params=params)
            if 'feed' in response.keys():
                for curr_entry in response['feed']['entry']:
                    time_start = np.append(time_start,np.datetime64(curr_entry['time_start'],'ns'))
                    for curr_link in curr_entry['links']:
                        if "direct download access via S3" in curr_link['title']:
                            s3_files_list.append(curr_link['href'])
                            break
            elif 'errors' in response.keys():
                raise Exception(response['errors'][0])
            
            if len(response['feed']['entry']) < 2000:
                completed_query = True
            else:
                # do another CMR search since previous search hit the allowed maximum
                # number of entries (2000)
                params['temporal'] = str(np.datetime64(response['feed']['entry'][-1]['time_end'],'D')\
                                         + np.timedelta64(1,'D'))+params['temporal'][10:]

        # reduce granule list to single day if only one day in requested range
        if (('MONTHLY' in ShortName) or ('DAILY' in ShortName)):
            if ((SingleDay_flag == True) and (len(s3_files_list) > 1)):
                day_index = np.argmin(np.abs(time_start - np.datetime64(StartDate,'D')))
                s3_files_list = s3_files_list[day_index:(day_index+1)]

        return s3_files_list
    
    
    def get_granules_ecco_bucket(StartDate: str, EndDate: str,\
                                   ShortName: str, version: str, SingleDay_flag: bool):
        import s3fs
        
        # find all granules in the dataset identified by ShortName
        s3 = s3fs.S3FileSystem(anon=False,\
                               requester_pays=True)
        if version == 'v4r5':
            if 'LLC0090GRID' in ShortName:
                gridtime_id = 'native/'
            elif '05DEG' in ShortName:
                gridtime_id = 'latlon/'
            if 'MONTHLY' in ShortName:
                gridtime_id += 'mon_mean/'
            elif 'DAILY' in ShortName:
                gridtime_id += 'day_mean/'
            elif 'SNAPSHOT' in ShortName:
                gridtime_id += 'snap/'
            shortname_dir = "_".join(ShortName.split("_")[2:-3])
            s3_files_all = s3.ls("s3://ecco-model-granules/netcdf/V4r5/"+gridtime_id\
                                 +shortname_dir+"/")
        
        # include only the granules in the date range given by temporal_range
        s3_files_all_datestr = [s3_file.split("_")[-5] for s3_file in s3_files_all]
        s3_files_all_dates = np.array([np.datetime64(datestr[:np.fmin(len(datestr),13)],'h')\
                                         for datestr in s3_files_all_datestr])
        sorted_ind = np.argsort(s3_files_all_dates)
        s3_files_all_dates = s3_files_all_dates[sorted_ind]
        
        in_range_ind = np.logical_and(\
                         s3_files_all_dates >= np.datetime64(StartDate,'D'),\
                         s3_files_all_dates <= np.datetime64(EndDate,'D'))\
                         .nonzero()[0]
        s3_files_list = [s3_files_all[sorted_ind[ind]] for ind in in_range_ind]

        # reduce granule list to single day if only one day in requested range
        if (('MONTHLY' in ShortName) or ('DAILY' in ShortName)):
            if ((SingleDay_flag == True) and (len(s3_files_list) > 1)):
                day_index = np.argmin(np.abs(time_start - np.datetime64(StartDate,'D')))
                s3_files_list = s3_files_list[day_index:(day_index+1)]
        
        
        return s3_files_list
    
    
    
    # # set default StartDate or EndDate if not previously provided
    if StartDate == None:
        StartDate = '1992-01-01'
    if EndDate == None:
        EndDate = '2099-12-31'
    
    # # Adjust StartDate and EndDate to CMR query values
    StartDate,EndDate,SingleDay_flag = date_adjustment(ShortName,\
                                         StartDate,EndDate,CMR_query=True)
    
    if version == 'v4r5':
        # Query ecco-model-granules S3 bucket for desired granules
        s3_files_list = get_granules_ecco_bucket(StartDate,EndDate,\
                                                 ShortName,version,SingleDay_flag)
    else:
        ## Log into Earthdata using your username and password
        setup_earthdata_login_auth()
        
        # Query the NASA Common Metadata Repository to find the URL of every granule associated with the desired 
        # ECCO Dataset and date range of interest.
        
        # create a Python dictionary with our search criteria:  `ShortName` and `temporal`
        input_search_params = {'ShortName': ShortName,
                               'temporal': ",".join([StartDate, EndDate])}
        
        print(input_search_params)
        
        # Query CMR for the desired ECCO Dataset
        s3_files_list = get_granules(input_search_params,ShortName,SingleDay_flag)
    
    # for snapshot datasets with monthly snapshot_interval, only include snapshots at beginning/end of months
    if 'SNAPSHOT' in ShortName:
        if snapshot_interval == 'monthly':
            import re
            s3_files_list_copy = list(tuple(s3_files_list))
            for s3_file in s3_files_list:
                snapshot_date = re.findall("_[0-9]{4}-[0-9]{2}-[0-9]{2}",s3_file)[0][1:]
                if snapshot_date[8:] != '01':
                    s3_files_list_copy.remove(s3_file)
            s3_files_list = s3_files_list_copy
    
    
    return s3_files_list



###================================================================================================================


def init_S3FileSystem(version):
    
    """
    
    This routine automatically pulls your EDL crediential from .netrc file and use it to obtain an AWS S3 credential 
    through a PO.DAAC service accessible at https://archive.podaac.earthdata.nasa.gov/s3credentials.
    From the PO.DAAC Github (https://podaac.github.io/tutorials/external/July_2022_Earthdata_Webinar.html).

    Parameters
    ----------
    version: ('v4r4','v4r5'), the ECCO version of the files.
    
    Returns:
    =======        
    s3: an AWS S3 filesystem
    
    """
    
    import s3fs
    
    if version == 'v4r5':
        s3 = s3fs.S3FileSystem(anon=False,requester_pays=True)
    else:
        creds = requests.get('https://archive.podaac.earthdata.nasa.gov/s3credentials').json()
        s3 = s3fs.S3FileSystem(anon=False,
                               key=creds['accessKeyId'],
                               secret=creds['secretAccessKey'], 
                               token=creds['sessionToken'])
    
    return s3



###================================================================================================================


def download_file(s3, url, output_dir, force, show_noredownload_msg):
    
    """
    Helper subroutine to gracefully download single files and avoids re-downloading if file already exists.
    To force redownload of the file, pass **True** to the boolean argument *force* (default **False**).

    Parameters
    ----------
    url: str, the HTTPS url from which the file will download
    output_dir: str, the local path into which the file will download
    force: bool, download even if the file exists locally already
    show_noredownload_msg (bool): show "no re-download" messages (vs. not showing messages)

    Returns
    -------
    target_file: str, downloaded file path
    
    """

    pass
    
    if not isdir(output_dir):
        raise Exception(f"Output directory doesn't exist! ({output_dir})")
    
    target_file = join(output_dir, basename(url))
    
    # if the file has already been downloaded, skip    
    if isfile(target_file) and force is False:
        if show_noredownload_msg:
            print(f'\n{basename(url)} already exists, and force=False, not re-downloading')
        return target_file

    # download file to local (output) file directory
    u_name = url.split('/')[-1]
    print(f'downloading {u_name}')
    s3.get_file(url, target_file)

    return target_file



###================================================================================================================


def download_files_concurrently(s3, dls, download_dir, n_workers, force=False, show_noredownload_msg=True):
    """Download files using thread pool with up to n_workers"""

    pass
    
    start_time = time.time()

    # use thread pool for concurrent downloads
    with ThreadPoolExecutor(max_workers=n_workers) as executor:

        # tqdm makes a cool progress bar
        downloaded_files = list(tqdm(executor.map(download_file, repeat(s3), dls, repeat(download_dir), repeat(force), repeat(show_noredownload_msg)),\
                                     total=len(dls), desc='DL Progress',\
                                     ascii=True, ncols=75, file=sys.stdout))
    
        # calculate total time spent in the download
        total_time_download = time.time() - start_time

        print('\n=====================================')
        print('Time spent = ' + str(total_time_download) + ' seconds')
        print('\n')

    return downloaded_files



###================================================================================================================


def download_files_s3_wrapper(s3, s3_files_list, download_dir, n_workers, force_redownload, show_noredownload_msg):
    """Wrapper for downloading functions"""

    pass
    
    try:
        # Attempt concurrent downloads, but if error arises switch to sequential downloads
        ### Method 1: Concurrent downloads        
        
        # Force redownload (or not) depending on value of force_redownload
        downloaded_files = download_files_concurrently(s3, s3_files_list, download_dir, n_workers, force_redownload, show_noredownload_msg)
        
    except:
        ### Method 2: Sequential Downloads
        
        start_time = time.time()
        
        # Download each URL sequentially in a for loop.
        total_download_size_in_bytes = 0
        
        # loop through all files
        downloaded_files = []
        for u in s3_files_list:
            result = download_file(s3, url=u, output_dir=download_dir, force=force_redownload,\
                                   show_noredownload_msg=show_noredownload_msg)
            downloaded_files.append(result)
        
        # calculate total time spent in the download
        total_time_download = time.time() - start_time
        
        print('\n=====================================')
        print('Time spent = ' + str(total_time_download) + ' seconds')
        print('\n')

        return downloaded_files



###================================================================================================================


def ecco_podaac_s3_open(ShortName,StartDate,EndDate,version,snapshot_interval='monthly',\
                        prompt_request_payer=True):
    
    """
    
    This routine searches for and opens ECCO datasets from S3 buckets in the PO.DAAC Cloud.
    It returns a list of opened file(s) on S3 that can be passed to xarray.
    This function is intended to be called from an EC2 instance running in AWS region us-west-2.
    
    Parameters
    ----------
    ShortName: str, the ShortName that identifies the dataset on PO.DAAC.
    
    StartDate,EndDate: str, in 'YYYY', 'YYYY-MM', or 'YYYY-MM-DD' format, 
                       define date range [StartDate,EndDate] for download.
                       EndDate is included in the time range (unlike typical Python ranges).
                       ECCOv4r4 date range is '1992-01-01' to '2017-12-31'.
                       For 'SNAPSHOT' datasets, an additional day is added to EndDate to enable closed budgets
                       within the specified date range.

    version: ('v4r4','v4r5'), specifies ECCO version to query.
             'v4r5' only has grid='native' and time_res='monthly' data files available.
    
    snapshot_interval: ('monthly', 'daily'), if the dataset corresponding to ShortName is a snapshot, 
                       determines whether snapshots are included for only the beginning/end of each month 
                       ('monthly'), or for every day ('daily'). Defaults to 'monthly'.
    
    prompt_request_payer: bool, if True (default), user is prompted to approve 
                                (by entering "y" or "Y") any access to a 
                                requester pays bucket, otherwise request is canceled; 
                                if False, data access proceeds without prompting.

    Returns
    -------
    open_files: str or list, opened file(s) on S3 that can be passed directly to xarray (open_dataset or open_mfdataset)
    
    """

    pass    
    
    
    # get list of files
    s3_files_list = ecco_podaac_s3_query(ShortName,StartDate,EndDate,version)

    num_grans = len(s3_files_list)
    print (f'\nTotal number of matching granules: {num_grans}')

    # initiate S3 access
    s3 = init_S3FileSystem(version)
    
    if ((version == 'v4r5') and prompt_request_payer):
        # give requester a chance to opt out of paying data transfer fees
        option_proceed = input("Files will be accessed from a requester pays S3 bucket.\n"\
                               +"Requester is responsible for any data transfer fees.\n"\
                               +"Do you want to proceed? [y/n]: ")
        if option_proceed.casefold() != 'y':
            raise Exception("Request canceled; no data transferred.")
    
    # open files and create list that can be passed to xarray file opener
    open_files = [s3.open(file) for file in s3_files_list]
    # if list has length 1, return a string instead of a list
    if len(open_files) == 1:
        open_files = open_files[0]

    return open_files



###================================================================================================================


def ecco_podaac_s3_open_fsspec(ShortName,version,jsons_root_dir=None,jsons_retrieve=True,\
                               prompt_request_payer=True):
    
    """
    
    This routine searches for and opens ECCO datasets from S3 buckets in the PO.DAAC Cloud.
    It returns a list of opened file(s) on S3 that can be passed to xarray.
    This function is intended to be called from an EC2 instance running in AWS region us-west-2.
    
    Parameters
    ----------
    ShortName: str, the ShortName that identifies the dataset on PO.DAAC.
    
    version: ('v4r4','v4r5'), specifies ECCO version to query.
    
    jsons_root_dir: str, the root/parent directory where the 
                    fsspec/kerchunk-generated jsons are found, 
                    or will be stored in if jsons_retrieve=True.
                    jsons are generated using the steps described here:
                    https://medium.com/pangeo/fake-it-until-you-make-it-reading-goes-netcdf4-data-on-aws-s3-as-zarr
                    -for-rapid-data-access-61e33f8fe685.
                    If None (default), jsons_root_dir will be set to ~/MZZ/{version}/.
                    The jsons need to be stored in the following directories/formats:
                    For v4r4: {jsons_root_dir}/MZZ_{GRIDTYPE}_{TIME_RES}/{SHORTNAME}.json.
                    GRIDTYPE is '05DEG' or 'LLC0090GRID' for v4r4;
                    TIME_RES is one of: ('MONTHLY','DAILY','SNAPSHOT','GEOMETRY','MIXING_COEFFS').
                    For v4r5: {jsons_root_dir}/MZZ_{TIME_RES}_{GRIDTYPE}/{SHORTNAME_core}_{TIME_RES}_{GRIDTYPE}_llc090_ECCOV4r5.json.
                    TIME_RES is one of: ('mon_mean','day_mean','snap');
                    GRIDTYPE is one of: ('native','latlon').
                    The SHORTNAME_core is the SHORTNAME without the 'ECCO_L4_' at the beginning, 
                    or the grid/time/version IDs at the end (e.g., 'LLC0090GRID_MONTHLY_V4R5').
    
    jsons_retrieve: bool, if True (default), and jsons are not already stored in jsons_root_dir, 
                    retrieve from S3 bucket and store locally in jsons_root_dir.
                    If False, and jsons are not stored in jsons_root_dir, 
                    the data retrieval will fail.
    
    prompt_request_payer: bool, if True (default), user is prompted to approve 
                                (by entering "y" or "Y") any access to a 
                                requester pays bucket, otherwise request is canceled; 
                                if False, data access proceeds without prompting.

    Returns
    -------
    fsmap_obj: fsspec.mapping.FSMap object, can be passed directly to xarray.open_dataset 
               (with engine='zarr')
    
    """

    pass
    
    import glob
    import fsspec
    
    
    # identify name of target json and local directory
    shortname_split = ShortName.split('_')
    if version == 'v4r4':
        if 'GEOMETRY' in ShortName:
            gridtype = shortname_split[-2]
            time_res = 'GEOMETRY'
            if gridtype == 'LLC0090GRID':
                json_basename = "GRID_GEOMETRY_ECCO_V4r4_native_llc0090.nc.json"
            elif gridtype == '05DEG':
                json_basename = "GRID_GEOMETRY_ECCO_V4r4_latlon_0p50deg.nc.json"
        elif 'MIX_COEFFS' in ShortName:
            gridtype = shortname_split[-2]
            time_res = 'MIXING_COEFFS'
            if gridtype == 'LLC0090GRID':
                json_basename = "OCEAN_3D_MIXING_COEFFS_ECCO_V4r4_native_llc0090.nc.json"
            elif gridtype == '05DEG':
                json_basename = "OCEAN_3D_MIXING_COEFFS_ECCO_V4r4_latlon_0p50deg.nc.json"
        else:
            gridtype = shortname_split[-3]
            time_res = shortname_split[-2]
            json_basename = ShortName+".json"
        json_subdir = "_".join(['MZZ',gridtype,time_res])
    elif version == 'v4r5':
        if 'MONTHLY' in ShortName:
            gridtime_id = 'mon_mean_'
        elif 'DAILY' in ShortName:
            gridtime_id = 'day_mean_'
        elif 'SNAPSHOT' in ShortName:
            gridtime_id = 'snap_'
        if 'LLC0090GRID' in ShortName:
            gridtime_id += 'native'
        elif '05DEG' in ShortName:
            gridtime_id += 'latlon'
        shortname_core = "_".join(ShortName.split("_")[2:-3])
        json_subdir = "MZZ_"+gridtime_id
        json_basename = shortname_core+"_"+gridtime_id+"_llc090_ECCOV4r5.json"
    
    if jsons_root_dir is None:
        json_local_subdir = join(expanduser('~'),'MZZ',version,json_subdir)
    else:
        json_local_subdir = join(jsons_root_dir,json_subdir)
    json_file = join(json_local_subdir,json_basename)
    
    if ((jsons_retrieve) and (not isfile(json_file))):
        # retrieve json from s3://ecco-model-granules, if it is not already stored locally
        import s3fs
        import os
        
        if prompt_request_payer:
            # give requester a chance to opt out of paying data transfer fees
            option_proceed = input("Files will be accessed from a requester pays S3 bucket.\n"\
                                   +"Requester is responsible for any data transfer fees.\n"\
                                   +"Do you want to proceed? [y/n]: ")
            if option_proceed.casefold() != 'y':
                raise Exception("Request canceled; no data transferred.")
        s3 = s3fs.S3FileSystem(anon=False,\
                               requester_pays=True)
         
        if version == 'v4r4':
            json_s3_subdir = "s3://ecco-model-granules/V4r4/mzz_jsons/" + json_subdir
        elif version == 'v4r5':
            json_s3_subdir = "s3://ecco-model-granules/netcdf/V4r5/MZZ/" + json_subdir
        json_s3_file = join(json_s3_subdir,json_basename)
        os.makedirs(json_local_subdir,exist_ok=True)
        s3.get_file(json_s3_file,json_file)
        
    elif not isfile(json_file):
        raise Exception("json file "+json_file+" not found.\n"\
                        +"To proceed, verify json file is stored in location above, \n"\
                        +"or set jsons_retrieve=True.")
    
    # access ECCO output using fsspec mapper object
    if version == 'v4r5':
        # generate mapper object
        fs = fsspec.filesystem(\
                    "reference",\
                    fo=json_file,\
                    remote_protocol="s3", 
                    remote_options={"anon":False,\
                                    "requester_pays":True})
    else:
        # get NASA Earthdata credentials for S3
        creds = requests.get('https://archive.podaac.earthdata.nasa.gov/s3credentials').json()
        
        # generate mapper object
        fs = fsspec.filesystem(\
                    "reference",\
                    fo=json_file,\
                    remote_protocol="s3", 
                    remote_options={"anon":False,\
                                    "key":creds['accessKeyId'],
                                    "secret":creds['secretAccessKey'], 
                                    "token":creds['sessionToken']})
    
    fs.asynchronous = True
    fsmap_obj = zarr.storage.FsspecStore(fs)
    
    return fsmap_obj



###================================================================================================================


def ecco_podaac_s3_get(ShortName,StartDate,EndDate,version,snapshot_interval='monthly',download_root_dir=None,\
                       n_workers=6,force_redownload=False,show_noredownload_msg=True,\
                       prompt_request_payer=True,\
                       return_downloaded_files=False):

    """
    
    This routine downloads ECCO datasets from PO.DAAC, to be stored locally on a AWS EC2 instance running in 
    region us-west-2. It is adapted from the ecco_podaac_download function in the ecco_download.py module, 
    and is the AWS Cloud equivalent of ecco_podaac_download.
    
    Parameters
    ----------
    
    ShortName: str, the ShortName that identifies the dataset on PO.DAAC.
    
    StartDate,EndDate: str, in 'YYYY', 'YYYY-MM', or 'YYYY-MM-DD' format, 
                       define date range [StartDate,EndDate] for download.
                       EndDate is included in the time range (unlike typical Python ranges).
                       ECCOv4r4 date range is '1992-01-01' to '2017-12-31'.
                       For 'SNAPSHOT' datasets, an additional day is added to EndDate to enable closed budgets
                       within the specified date range.
    version: ('v4r4','v4r5'), specifies ECCO version to query.
             'v4r5' only has grid='native' and time_res='monthly' data files available.
    
    snapshot_interval: ('monthly', 'daily'), if the dataset corresponding to ShortName is a snapshot, 
                       determines whether snapshots are included for only the beginning/end of each month 
                       ('monthly'), or for every day ('daily'). Defaults to 'monthly'.
    
    download_root_dir: str, defines parent directory to download files to.
                       Files will be downloaded to directory download_root_dir/ShortName/.
                       If not specified, parent directory defaults to '~/Downloads/ECCO_V4r4_PODAAC/',
                       or '~/Downloads/ECCO_V4r5_PODAAC/' if version == 'v4r5'.
    
    n_workers: int, number of workers to use in concurrent downloads. Benefits typically taper off above 5-6.
    
    force_redownload: bool, if True, existing files will be redownloaded and replaced;
                            if False, existing files will not be replaced.
    
    show_noredownload_msg: bool, if True (default), and force_redownload=False, 
                               display message for each file that is already 
                               downloaded (and therefore not re-downloaded); 
                               if False, these messages are not shown.
    
    prompt_request_payer: bool, if True (default), user is prompted to approve 
                                (by entering "y" or "Y") any access to a 
                                requester pays bucket, otherwise request is canceled; 
                                if False, data access proceeds without prompting.
    
    return_downloaded_files: bool, if True, string or list of downloaded file(s) (including files that were 
                             already on disk and not replaced) is returned.
                             If False (default), the function returns nothing.

    Returns
    -------
    downloaded_files: str or list, downloaded file(s) with local path that can be passed 
                      directly to xarray (open_dataset or open_mfdataset).
                      Only returned if return_downloaded_files=True.
    
    """

    pass
    
    from concurrent.futures import ThreadPoolExecutor
 

    # set default download parent directory
    if download_root_dir==None:
        if version == 'v4r4':
            download_root_dir = join(expanduser('~'),'Downloads','ECCO_V4r4_PODAAC')
        elif version == 'v4r5':
            download_root_dir = join(expanduser('~'),'Downloads','ECCO_V4r5_PODAAC')

    # define the directory where the downloaded files will be saved
    download_dir = Path(download_root_dir) / ShortName
    
    # create the download directory if it does not already exist
    if isdir(download_dir) == True:
        print(f'Download to directory {download_dir}')
    else:
        print(f'Creating download directory {download_dir}')
    download_dir.mkdir(exist_ok = True, parents=True)
    
    # get list of files
    s3_files_list = ecco_podaac_s3_query(ShortName,StartDate,EndDate,version)
    
    num_grans = len(s3_files_list)
    print (f'\nTotal number of matching granules: {num_grans}')

    # initiate S3 access
    s3 = init_S3FileSystem(version)

    if ((version == 'v4r5') and prompt_request_payer):
        # give requester a chance to opt out of paying data transfer fees
        option_proceed = input("Files will be accessed from a requester pays S3 bucket.\n"\
                               +"Requester is responsible for data transfer fees.\n"\
                               +"Do you want to proceed? [y/n]: ")
        if option_proceed.casefold() != 'y':
            raise Exception("Request canceled; no data transferred.")
    
    # download files
    downloaded_files = download_files_s3_wrapper(s3, s3_files_list, download_dir, n_workers, force_redownload, show_noredownload_msg)
    
    if return_downloaded_files == True:
        if len(downloaded_files) == 1:
            # if only 1 file is downloaded, return a string of filename instead of a list
            downloaded_files = downloaded_files[0]
        return downloaded_files



###================================================================================================================


def ecco_podaac_s3_get_diskaware(ShortNames,StartDate,EndDate,version,snapshot_interval=None,\
                                 download_root_dir=None,max_avail_frac=0.5,n_workers=6,force_redownload=False,\
                                 show_noredownload_msg=True,prompt_request_payer=True):
    
    """
    
    This function estimates the storage footprint of ECCO datasets, given ShortName(s), a date range, and which 
    files (if any) are already present.
    If the footprint of the files to be downloaded (not including files already on the instance or re-downloads) 
    is <= the max_avail_frac specified of the instance's available storage, they are downloaded and stored locally 
    on the instance (hosting files locally typically speeds up loading and computation).
    Otherwise, the files are "opened" using ecco_podaac_s3_open so that they can be accessed directly 
    on S3 without occupying local storage.

    Parameters
    ----------
    
    ShortNames: str or list, the ShortName(s) that identify the dataset on PO.DAAC.
    
    StartDate,EndDate: str, in 'YYYY', 'YYYY-MM', or 'YYYY-MM-DD' format, 
                       define date range [StartDate,EndDate] for download.
                       EndDate is included in the time range (unlike typical Python ranges).
                       ECCOv4r4 date range is '1992-01-01' to '2017-12-31'.
                       For 'SNAPSHOT' datasets, an additional day is added to EndDate to enable closed budgets
                       within the specified date range.
    
    version: ('v4r4','v4r5'), specifies ECCO version to query.
             'v4r5' only has grid='native' and time_res='monthly' data files available.
    
    snapshot_interval: ('monthly', 'daily', or None), if snapshot datasets are included in ShortNames, 
                       this determines whether snapshots are included for only the beginning/end of each month 
                       ('monthly'), or for every day ('daily').
                       If None or not specified, defaults to 'daily' if any daily mean ShortNames are included 
                       and 'monthly' otherwise.

    download_root_dir: str, defines parent directory to download files to.
                       Files will be downloaded to directory download_root_dir/ShortName/.
                       If not specified, parent directory defaults to '~/Downloads/ECCO_V4r4_PODAAC/',
                       or '~/Downloads/ECCO_V4r5_PODAAC/' if version == 'v4r5'.
    
    max_avail_frac: float, maximum fraction of remaining available disk space to use in storing current ECCO datasets.
                    This determines whether the dataset files are stored on the current instance, or opened on S3.
                    Valid range is [0,0.9]. If number provided is outside this range, it is replaced by the closer 
                    endpoint of the range.
    
    n_workers: int, number of workers to use in concurrent downloads. Benefits typically taper off above 5-6.
               Applies only if files are downloaded.
    
    force_redownload: bool, if True, existing files will be redownloaded and replaced;
                            if False, existing files will not be replaced.
                            Applies only if files are downloaded.
    
    show_noredownload_msg: bool, if True (default), and force_redownload=False, 
                               display message for each file that is already 
                               downloaded (and therefore not re-downloaded); 
                               if False, these messages are not shown.
    
    prompt_request_payer: bool, if True (default), user is prompted to approve 
                                (by entering "y" or "Y") any access to a 
                                requester pays bucket, otherwise request is canceled; 
                                if False, data access proceeds without prompting.
    
    
    Returns
    -------
    retrieved_files: dict, with keys: ShortNames and values: downloaded or opened file(s) with path on local instance 
                     or on S3, that can be passed directly to xarray (open_dataset or open_mfdataset).
    
    """

    pass

    import shutil
    
    
    # force max_avail_frac to be within limits [0,0.9]
    max_avail_frac = np.fmin(np.fmax(max_avail_frac,0),0.9)
    
    # initiate S3 access
    s3 = init_S3FileSystem(version)

    # determine value of snapshot_interval if None or not specified
    if snapshot_interval == None:
        snapshot_interval = 'monthly'
        for curr_shortname in ShortNames:
            if 'DAILY' in curr_shortname:
                snapshot_interval = 'daily'
                break

    # set default download parent directory
    if download_root_dir==None:
        if version == 'v4r4':
            download_root_dir = join(expanduser('~'),'Downloads','ECCO_V4r4_PODAAC')
        elif version == 'v4r5':
            download_root_dir = join(expanduser('~'),'Downloads','ECCO_V4r5_PODAAC')

    # add up total size of files that would be downloaded
    dataset_sizes = np.array([])
    s3_files_list_all = []
    for curr_shortname in ShortNames:
        
        # get list of files
        s3_files_list = ecco_podaac_s3_query(curr_shortname,StartDate,EndDate,version,snapshot_interval)
        
        # create the download directory if it does not already exist
        download_dir = Path(download_root_dir) / curr_shortname
        if isdir(download_dir) == True:
            print(f'Download to directory {download_dir}')
        else:
            print(f'Creating download directory {download_dir}')
        download_dir.mkdir(exist_ok = True, parents=True)
        
        # compute size of current dataset
        curr_dataset_size = 0
        for s3_file in s3_files_list:
            if isfile(join(download_dir,basename(s3_file))) == False:
                curr_dataset_size += s3.info(s3_file)['size']

        dataset_sizes = np.append(dataset_sizes,curr_dataset_size)
        s3_files_list_all.append(s3_files_list)
            

    # query available disk space at download location
    query_disk_completed = False
    query_dir = [download_root_dir][0]
    while query_disk_completed == False:
        try:
            avail_storage = shutil.disk_usage(query_dir).free
            query_disk_completed = True
        except:
            try:
                query_dir = join(*os.path.split(query_dir)[:-1])
            except:                
                print('Error: can not detect available disk space for download_root_dir: '+download_root_dir)
                return -1

    # fraction of available storage that would be occupied by downloads
    sizes_sum = np.sum(dataset_sizes)
    avail_frac = sizes_sum/avail_storage

    print(f'Size of files to be downloaded to instance is {(1.e-3)*np.round((1.e3)*sizes_sum/(2**30))} GB,\n'\
                +f'which is {.01*np.round((1.e4)*avail_frac)}% of the {(1.e-3)*np.round((1.e3)*avail_storage/(2**30))} GB available storage.')
    
    if ((version == 'v4r5') and prompt_request_payer):
        # give requester a chance to opt out of paying data transfer fees
        option_proceed = input("Files will be accessed from a requester pays S3 bucket.\n"\
                               +"Requester is responsible for data transfer fees.\n"\
                               +"Do you want to proceed? [y/n]: ")
        if option_proceed.casefold() != 'y':
            raise Exception("Request canceled; no data transferred.")

    retrieved_files = {}
    if avail_frac <= max_avail_frac:
        # proceed with file downloads
        print('Proceeding with downloads of any needed files from S3.')
        for curr_shortname,s3_files_list in zip(ShortNames,s3_files_list_all):
            # set default download parent directory
            if download_root_dir==None:
                if version == 'v4r4':
                    download_root_dir = join(expanduser('~'),'Downloads','ECCO_V4r4_PODAAC')
                elif version == 'v4r5':
                    download_root_dir = join(expanduser('~'),'Downloads','ECCO_V4r5_PODAAC')
        
            # define the directory where the downloaded files will be saved
            download_dir = Path(download_root_dir) / curr_shortname
            
            # download files
            downloaded_files = download_files_s3_wrapper(s3, s3_files_list, download_dir, n_workers, force_redownload, show_noredownload_msg)

            if len(downloaded_files) == 1:
                # if only 1 file is downloaded, return a string of filename instead of a list
                downloaded_files = downloaded_files[0]

            retrieved_files[curr_shortname] = downloaded_files

    else:
        # open files from S3 instead of downloading
        print('Download size is larger than specified fraction of available storage.\n'\
              +'Generating file lists to open directly from S3.')

        for curr_shortname,s3_files_list in zip(ShortNames,s3_files_list_all):
            # open files and create list that can be passed to xarray file opener
            open_files = [s3.open(file) for file in s3_files_list]
            # if list has length 1, return a string instead of a list
            if len(open_files) == 1:
                open_files = open_files[0]

            retrieved_files[curr_shortname] = open_files

    return retrieved_files



"""ECCO v4 Python: read_bin_gen

This module includes utility routines for loading general binary files,
not necessarily in the lat-lon-cap layout. For LLC type data, see read_bin_llc.

.. _ecco_v4_py Documentation :
   https://github.com/ECCO-GROUP/ECCOv4-py
"""

from __future__ import division,print_function
import numpy as np
import glob
import os
from pathlib import Path

#%%
def load_binary_array(fdir, fname, ni, nj, nk=1, nl=1, skip=0,
                      filetype = '>f', less_output = False ):
    """
    Note: This function is for reading a general binary file. To read data in
    the llc structure, see read_bin_llc.

    Loads a binary array from a file into memory, which is not necessarily in
    llc format.  The first two dimensions
    of the array have length ni and nj, respectively.  The array is comprised
    of one or more 2D 'slices' of dimension ni x nj.  The number of 2D slices
    to read is 'nk'.  nk is not necessarily the length of the third dimension
    of the file.  The 'skip' parameter specifies the number of 2D slices
    to skip over before reading the nk number of 2D slices.  nk can be 1.

    Parameters
    ----------
    fdir : string
        A string with the directory of the binary file to open
    fname : string
        A string with the name of the binary file to open
    ni,nj : int
        the length of each array dimension.  ni, nj must be > 0
    skip : int
        the number of 2D (nj x ni) slices to skip.
        Default: 0
    nk : int
        number of 2D slices (or records) to load in the third dimension.
        if nk = -1, load all 2D slices
        Default: 1 [singleton]
    nl : int
        number of 2D slices (or records) to load in the fourth dimension.
        Default: 1 [singleton]
    filetype: string
        the file type, default is big endian (>) 32 bit float (f)
        alternatively, ('<d') would be little endian (<) 64 bit float (d)
        Default '>f'
    less_output : boolean
        A debugging flag.  False = less debugging output
        Default: False

    Returns
    -------
    data
        a numpy array with dimensions nl x nk x nj x ni

    Raises
    ------
    IOError
        If the file is not found

    """
    #datafile = fdir + '/' + fname
    data_folder = Path(fdir)
    datafile = data_folder / fname

    #datafile = os.path.join(fdir, fname)
    if not less_output:
        print('load_binary_array: loading file', datafile)

    # check to see if file exists.
    if datafile.exists() == False:
        raise IOError(fname + ' not found ')

    f = open(datafile, 'rb')
    dt = np.dtype(filetype)

    if skip > 0:
        # skip ahead 'skip' number of 2D slices
        f.seek(ni*nj*skip*dt.itemsize)

    if (ni <= 0) or (nj <= 0):
        raise TypeError('ni and nj must be > 1')

    # load all 2D records
    if nk == -1:
        # nl can only be 1 if we use nk = -1
        nl = 1
        # read all 2D records
        arr_k = np.fromfile(f, dtype=filetype, count=-1)
        # find the number of 2D records (nk)
        length_arr_k = len(arr_k)

        # length of each 2D slice is ni * nj
        nk = int(length_arr_k / (ni*nj))

        if not less_output:
            print('load_binary_array: loading all 2D records.  nk =',nk)

        # reshape the array to 2D records
        if nk > 1: # we have more than one 2D record, make 3D field
            data = np.reshape(arr_k,(nk, nj, ni))

        else: # nk = 1, just make 2D field
            data = np.reshape(arr_k,(nj, ni))

    # read a specific number of records (nk*nl)
    else:
        if (nk <= 0) or (nl <= 0):
            raise TypeError('nk and nl must be > 0.  If they are singleton dimensions, use 1')

        # read in nk*nl 2D records
        arr_k = np.fromfile(f, dtype=filetype, count=ni*nj*nk*nl)

        # put data into 2D records
        #  - if we have a fourth dimension
        if nl > 1:
            data = np.reshape(arr_k,(nl, nk, nj, ni))

        #  - if we have a third dimension
        elif nk > 1:
            data = np.reshape(arr_k,(nk, nj, ni))

        #  - if we only have two dimensions
        else:
            data = np.reshape(arr_k,(nj, ni))

    f.close()

    if not less_output:
        print('load_binary_array: data array shape ', data.shape)
        print('load_binary_array: data array type ', data.dtype)

    return data



"""
Module for computing meridional overturning streamfunctions
TBD: add barotropic streamfunction
"""

import numpy as np

from .ecco_utils import get_llc_grid
from .calc_meridional_trsp import _parse_coords, meridional_trsp_at_depth
from .calc_section_trsp import _parse_section_trsp_inputs, section_trsp_at_depth

# Define constants
METERS_CUBED_TO_SVERDRUPS = 10**-6

def calc_meridional_stf(ds,lat_vals,doFlip=True,
                        basin_name=None,coords=None,grid=None):
    """Compute the meridional overturning streamfunction in Sverdrups
    at specified latitude(s)

    Parameters
    ----------
    ds : xarray DataSet
        must contain UVELMASS,VVELMASS, drF, dyG, dxG
    lat_vals : float or list
        latitude value(s) rounded to the nearest degree
        specifying where to compute overturning streamfunction
    doFlip : logical, optional
        True: integrate from "bottom" by flipping Z dimension before cumsum(),
        then multiply by -1. False: flip neither dim nor sign.
    basin_name : string, optional
        denote ocean basin over which to compute streamfunction
        If not specified, compute global quantity
        see utils.get_available_basin_names for options
    coords : xarray Dataset
        separate dataset containing the coordinate information
        YC, drF, dyG, dxG, optionally maskW, maskS
    grid : xgcm Grid object, optional
        denotes LLC90 operations for xgcm, see ecco_utils.get_llc_grid
        see also the [xgcm documentation](https://xgcm.readthedocs.io/en/latest/grid_topology.html)

    Returns
    -------
    ds_out : xarray Dataset
        with the following variables
            moc
                meridional overturning strength as maximum of streamfunction
                in depth space, with dimensions 'time' (if in dataset), and 'lat'
            psi_moc
                meridional overturning streamfunction across the section in Sv
                with dimensions 'time' (if in given dataset), 'lat', and 'k'
            trsp_z
                freshwater transport across section at each depth level in Sv
                with dimensions 'time' (if in given dataset), 'lat', and 'k'
    """

    # get coords
    coords = _parse_coords(ds,coords,['Z','YC','drF','dyG','dxG'])

    # Compute volume transport
    trsp_x = ds['UVELMASS'] * coords['drF'] * coords['dyG']
    trsp_y = ds['VVELMASS'] * coords['drF'] * coords['dxG']

    # Creates an empty streamfunction
    ds_out = meridional_trsp_at_depth(trsp_x, trsp_y,
                                      lat_vals=lat_vals,
                                      coords=coords,
                                      basin_name=basin_name,
                                      grid=grid)

    psi_moc = ds_out['trsp_z'].copy(deep=True)

    # Flip depth dimension, take cumulative sum, flip back
    if doFlip:
        psi_moc = psi_moc.isel(k=slice(None,None,-1))

    # Should this be done with a grid object???
    psi_moc = psi_moc.cumsum(dim='k')

    if doFlip:
        psi_moc = -1 * psi_moc.isel(k=slice(None,None,-1))

    # Add to dataset
    ds_out['psi_moc'] = psi_moc

    # Compute overturning strength
    ds_out['moc'] = ds_out['psi_moc'].max(dim='k')

    # Convert to Sverdrups
    for fld in ['trsp_z','psi_moc','moc']:
        ds_out[fld] = ds_out[fld] * METERS_CUBED_TO_SVERDRUPS
        ds_out[fld].attrs['units'] = 'Sv'

    # Name the fields here, after unit conversion which doesn't keep attrs
    ds_out['trsp_z'].attrs['name'] = 'volumetric trsp per depth level'
    ds_out['psi_moc'].attrs['name'] = 'meridional overturning streamfunction'
    ds_out['moc'].attrs['name'] = 'meridional overturning strength'

    return ds_out

def calc_section_stf(ds,
                     pt1=None, pt2=None,
                     section_name=None,
                     maskW=None, maskS=None,
                     doFlip=True,coords=None,grid=None):
    """Compute the overturning streamfunction in plane normal to
    section defined by pt1 and pt2 in depth space

    See calc_section_trsp.calc_section_vol_trsp for the various ways
    to call this function

    All inputs are the same except:

    Parameters
    ----------
    doFlip : logical, optional
        True: integrate from "bottom" by flipping Z dimension before cumsum(),
        then multiply by -1. False: flip neither dim nor sign.

    Returns
    -------
    ds_out : xarray Dataset
        with the following variables
            moc
                meridional overturning strength as maximum of streamfunction
                in depth space, with dimensions 'time' (if in dataset), and 'lat'
            psi_moc
                overturning streamfunction across the section in Sv
                with dimensions 'time' (if in given dataset), 'lat', and 'k'
            trsp_z
                freshwater transport across section at each depth level in Sv
                with dimensions 'time' (if in given dataset), 'lat', and 'k'
            maskW, maskS
                defining the section
        and the section_name as an attribute if it is provided
    """

    coords = _parse_coords(ds,coords,['Z','YC','XC','drF','dyG','dxG'])

    # Compute volume transport
    trsp_x = ds['UVELMASS'] * coords['drF'] * coords['dyG']
    trsp_y = ds['VVELMASS'] * coords['drF'] * coords['dxG']

    maskW, maskS = _parse_section_trsp_inputs(coords,pt1,pt2,maskW,maskS,section_name,
                                              grid=grid)

    # Creates an empty streamfunction
    ds_out = section_trsp_at_depth(trsp_x, trsp_y,
                                   maskW, maskS,
                                   coords=coords)

    psi_moc = ds_out['trsp_z'].copy(deep=True)

    # Flip depth dimension, take cumulative sum, flip back
    if doFlip:
        psi_moc = psi_moc.isel(k=slice(None,None,-1))

    # Should this be done with a grid object???
    psi_moc = psi_moc.cumsum(dim='k')

    if doFlip:
        psi_moc = -1 * psi_moc.isel(k=slice(None,None,-1))

    # Add to dataset
    ds_out['psi_moc'] = psi_moc

    # Compute overturning strength
    ds_out['moc'] = ds_out['psi_moc'].max(dim='k')

    # Convert to Sverdrups
    for fld in ['trsp_z','psi_moc','moc']:
        ds_out[fld] = ds_out[fld] * METERS_CUBED_TO_SVERDRUPS
        ds_out[fld].attrs['units'] = 'Sv'

    # Name the fields here, after unit conversion which doesn't keep attrs
    ds_out['trsp_z'].attrs['name'] = 'volumetric trsp per depth level'
    ds_out['psi_moc'].attrs['name'] = 'overturning streamfunction'
    ds_out['moc'].attrs['name'] = 'overturning strength'

    # Add section name and masks to Dataset
    ds_out['maskW'] = maskW
    ds_out['maskS'] = maskS
    if section_name is not None:
        ds_out.attrs['name'] = section_name

    return ds_out
